from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _, ngettext
from rest_framework import serializers
from rest_framework.response import Response

from ....acl.objectacl import add_acl_to_obj
from ....conf import settings
from ....core.apipatch import ApiPatch
from ...models import PostLike
from ...moderation import posts as moderation
from ...permissions import (
    allow_approve_post,
    allow_hide_best_answer,
    allow_hide_post,
    allow_protect_post,
    allow_unhide_post,
    exclude_invisible_posts,
)

post_patch_dispatcher = ApiPatch()


def patch_acl(request, post, value):
    """useful little op that updates post acl to current state"""
    if value:
        add_acl_to_obj(request.user_acl, post)
        return {"acl": post.acl}
    return {"acl": None}


post_patch_dispatcher.add("acl", patch_acl)


def patch_is_liked(request, post, value):
    if not post.acl["can_like"]:
        raise PermissionDenied(_("You can't like posts in this category."))

    # lock user to protect us from likes flood
    request.user.lock()

    # grab like state for this post and user
    try:
        user_like = post.postlike_set.get(liker=request.user)
    except PostLike.DoesNotExist:
        user_like = None

    # no change
    if (value and user_like) or (not value and not user_like):
        return {
            "likes": post.likes,
            "last_likes": post.last_likes or [],
            "is_liked": value,
        }

    # like
    if value:
        post.postlike_set.create(
            category=post.category,
            thread=post.thread,
            liker=request.user,
            liker_name=request.user.username,
            liker_slug=request.user.slug,
        )
        post.likes += 1

    # unlike
    if not value:
        user_like.delete()
        post.likes -= 1

    post.last_likes = []
    for like in post.postlike_set.all()[:4]:
        post.last_likes.append({"id": like.liker_id, "username": like.liker_name})

    post.save(update_fields=["likes", "last_likes"])

    return {"likes": post.likes, "last_likes": post.last_likes or [], "is_liked": value}


post_patch_dispatcher.replace("is-liked", patch_is_liked)


def patch_is_protected(request, post, value):
    allow_protect_post(request.user_acl, post)
    if value:
        moderation.protect_post(request.user, post)
    else:
        moderation.unprotect_post(request.user, post)
    return {"is_protected": post.is_protected}


post_patch_dispatcher.replace("is-protected", patch_is_protected)


def patch_is_unapproved(request, post, value):
    allow_approve_post(request.user_acl, post)

    if value:
        raise PermissionDenied(_("Content approval can't be reversed."))

    moderation.approve_post(request.user, post)

    return {"is_unapproved": post.is_unapproved}


post_patch_dispatcher.replace("is-unapproved", patch_is_unapproved)


def patch_is_hidden(request, post, value):
    if value is True:
        allow_hide_post(request.user_acl, post)
        allow_hide_best_answer(request.user_acl, post)
        moderation.hide_post(request.user, post)
    elif value is False:
        allow_unhide_post(request.user_acl, post)
        moderation.unhide_post(request.user, post)

    return {"is_hidden": post.is_hidden}


post_patch_dispatcher.replace("is-hidden", patch_is_hidden)


def post_patch_endpoint(request, post):
    old_is_unapproved = post.is_unapproved

    response = post_patch_dispatcher.dispatch(request, post)

    # diff posts's state against pre-patch and resync thread/category if necessarys
    if old_is_unapproved != post.is_unapproved:
        post.thread.synchronize()
        post.thread.save()

        post.category.synchronize()
        post.category.save()

    return response


def bulk_patch_endpoint(request, thread):
    serializer = BulkPatchSerializer(
        data=request.data, context={"settings": request.settings}
    )
    if not serializer.is_valid():
        return Response(serializer.errors, status=400)

    posts = clean_posts_for_patch(request, thread, serializer.data["ids"])

    old_unapproved_posts = [p.is_unapproved for p in posts].count(True)

    response = post_patch_dispatcher.dispatch_bulk(request, posts)

    new_unapproved_posts = [p.is_unapproved for p in posts].count(True)

    if old_unapproved_posts != new_unapproved_posts:
        thread.synchronize()
        thread.save()

        thread.category.synchronize()
        thread.category.save()

    return response


def clean_posts_for_patch(request, thread, posts_ids):
    posts_queryset = exclude_invisible_posts(
        request.user_acl, thread.category, thread.post_set
    )
    posts_queryset = posts_queryset.filter(id__in=posts_ids, is_event=False).order_by(
        "id"
    )

    posts = []
    for post in posts_queryset:
        post.category = thread.category
        post.thread = thread
        posts.append(post)

    if len(posts) != len(posts_ids):
        raise PermissionDenied(_("One or more posts to update could not be found."))

    return posts


class BulkPatchSerializer(serializers.Serializer):
    ids = serializers.ListField(
        child=serializers.IntegerField(min_value=1), min_length=1
    )
    ops = serializers.ListField(
        child=serializers.DictField(), min_length=1, max_length=10
    )

    def validate_ids(self, data):
        settings = self.context["settings"]
        limit = settings.posts_per_page + settings.posts_per_page_orphans
        if len(data) > limit:
            message = ngettext(
                "No more than %(limit)s post can be updated at a single time.",
                "No more than %(limit)s posts can be updated at a single time.",
                limit,
            )
            raise serializers.ValidationError(message % {"limit": limit})
        return data



import time

from django.core.management.base import BaseCommand

from ....core.management.progressbar import show_progress
from ....core.pgutils import chunk_queryset
from ...models import Thread


class Command(BaseCommand):
    help = "Synchronizes threads"

    def handle(self, *args, **options):
        threads_to_sync = Thread.objects.count()

        if not threads_to_sync:
            self.stdout.write("\n\nNo threads were found")
        else:
            self.sync_threads(threads_to_sync)

    def sync_threads(self, threads_to_sync):
        self.stdout.write("Synchronizing %s threads...\n" % threads_to_sync)

        synchronized_count = 0
        show_progress(self, synchronized_count, threads_to_sync)
        start_time = time.time()

        for thread in chunk_queryset(Thread.objects.all()):
            thread.synchronize()
            thread.save()

            synchronized_count += 1
            show_progress(self, synchronized_count, threads_to_sync, start_time)

        self.stdout.write("\n\nSynchronized %s threads" % synchronized_count)



# Generated by Django 1.11.9 on 2018-03-26 00:10
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_threads", "0008_auto_20180310_2234")]

    operations = []



import pytest

from ...conf import settings
from ..providers import PermissionProviders


def test_providers_are_not_loaded_on_container_init():
    providers = PermissionProviders()

    assert not providers._initialized
    assert not providers._providers
    assert not providers._annotators
    assert not providers._user_acl_serializers


def test_container_loads_providers():
    providers = PermissionProviders()
    providers.load()

    assert providers._providers
    assert providers._annotators
    assert providers._user_acl_serializers


def test_loading_providers_second_time_raises_runtime_error():
    providers = PermissionProviders()
    providers.load()

    with pytest.raises(RuntimeError):
        providers.load()


def test_container_returns_list_of_providers():
    providers = PermissionProviders()
    providers.load()

    providers_setting = settings.MISAGO_ACL_EXTENSIONS
    assert len(providers.list()) == len(providers_setting)


def test_container_returns_dict_of_providers():
    providers = PermissionProviders()
    providers.load()

    providers_setting = settings.MISAGO_ACL_EXTENSIONS
    assert len(providers.dict()) == len(providers_setting)


def test_accessing_providers_list_before_load_raises_assertion_error():
    providers = PermissionProviders()
    with pytest.raises(AssertionError):
        providers.list()


def test_accessing_providers_dict_before_load_raises_assertion_error():
    providers = PermissionProviders()
    with pytest.raises(AssertionError):
        providers.dict()


def test_getter_returns_registered_type_annotator():
    class TestType:
        pass

    def test_annotator():
        pass

    providers = PermissionProviders()
    providers.acl_annotator(TestType, test_annotator)
    providers.load()

    assert test_annotator in providers.get_obj_type_annotators(TestType())


def test_container_returns_list_of_user_acl_serializers():
    providers = PermissionProviders()
    providers.load()

    assert providers.get_user_acl_serializers()


def test_getter_returns_registered_user_acl_serializer():
    def test_user_acl_serializer():
        pass

    providers = PermissionProviders()
    providers.user_acl_serializer(test_user_acl_serializer)
    providers.load()

    assert test_user_acl_serializer in providers.get_user_acl_serializers()



from django import forms
from django.contrib import messages
from django.utils.translation import gettext, gettext_lazy as _

from ....admin.forms import YesNoSwitch
from .base import ChangeSettingsForm

OAUTH2_OPTIONAL_FIELDS = (
    "oauth2_token_extra_headers",
    "oauth2_user_extra_headers",
    "oauth2_send_welcome_email",
    "oauth2_json_avatar_path",
)


class ChangeOAuth2SettingsForm(ChangeSettingsForm):
    settings = [
        "enable_oauth2_client",
        "oauth2_provider",
        "oauth2_client_id",
        "oauth2_client_secret",
        "oauth2_scopes",
        "oauth2_login_url",
        "oauth2_token_url",
        "oauth2_token_extra_headers",
        "oauth2_json_token_path",
        "oauth2_user_url",
        "oauth2_user_method",
        "oauth2_user_token_location",
        "oauth2_user_token_name",
        "oauth2_user_extra_headers",
        "oauth2_send_welcome_email",
        "oauth2_json_id_path",
        "oauth2_json_name_path",
        "oauth2_json_email_path",
        "oauth2_json_avatar_path",
    ]

    enable_oauth2_client = YesNoSwitch(
        label=_("Enable OAuth2 client"),
        help_text=_(
            "Enabling OAuth2 will make login option redirect users to the OAuth provider "
            "configured below. It will also disable option to register on forum, "
            "change username, email or password, as those features will be delegated "
            "to the 3rd party site."
        ),
    )
    oauth2_provider = forms.CharField(
        label=_("Provider name"),
        help_text=_("Name of the OAuth 2 provider to be displayed by interface."),
        max_length=255,
        required=False,
    )
    oauth2_client_id = forms.CharField(
        label=_("Client ID"),
        max_length=200,
        required=False,
    )
    oauth2_client_secret = forms.CharField(
        label=_("Client Secret"),
        max_length=200,
        required=False,
    )
    oauth2_scopes = forms.CharField(
        label=_("Scopes"),
        help_text=_("List of scopes to request from provider, separated with spaces."),
        max_length=500,
        required=False,
    )

    oauth2_login_url = forms.URLField(
        label=_("Login form URL"),
        help_text=_(
            "Address to login form on provider's server that users will be "
            "redirected to."
        ),
        max_length=500,
        required=False,
    )

    oauth2_token_url = forms.URLField(
        label=_("Access token retrieval URL"),
        help_text=_(
            "URL that will be called after user completes the login process "
            "and authorization code is sent back to your site. This URL "
            "is expected to take this code and return the access token that "
            "will be next used to retrieve user data."
        ),
        max_length=500,
        required=False,
    )
    oauth2_token_extra_headers = forms.CharField(
        label=_("Extra HTTP headers in token request"),
        help_text=_(
            "List of extra headers to include in a HTTP request made to retrieve "
            'the access token. Example header is "Header-name: value". Specify each '
            "header on separate line."
        ),
        widget=forms.Textarea(attrs={"rows": 4}),
        max_length=500,
        required=False,
    )
    oauth2_json_token_path = forms.CharField(
        label=_("JSON path to access token"),
        help_text=_(
            "Name of key containing the access token in JSON returned by the provider "
            'If token is nested, use period (".") for path, eg: "result.token" '
            'will retrieve the token from "token" key nested in "result".'
        ),
        max_length=500,
        required=False,
    )

    oauth2_user_url = forms.URLField(
        label=_("User data URL"),
        max_length=500,
        required=False,
    )
    oauth2_user_method = forms.ChoiceField(
        label=_("Request method"),
        choices=[
            ("POST", "POST"),
            ("GET", "GET"),
        ],
        widget=forms.RadioSelect(),
    )
    oauth2_user_token_location = forms.ChoiceField(
        label=_("Access token location"),
        choices=[
            ("QUERY", _("Query string")),
            ("HEADER", _("HTTP header")),
            ("HEADER_BEARER", _("HTTP header (Bearer)")),
        ],
        widget=forms.RadioSelect(),
    )
    oauth2_user_token_name = forms.CharField(
        label=_("Access token name"),
        max_length=200,
        required=False,
    )
    oauth2_user_extra_headers = forms.CharField(
        label=_("Extra HTTP headers in user request"),
        help_text=_(
            "List of extra headers to include in a HTTP request made to retrieve "
            'the user profile. Example header is "Header-name: value". Specify each '
            "header on separate line."
        ),
        widget=forms.Textarea(attrs={"rows": 4}),
        max_length=500,
        required=False,
    )

    oauth2_send_welcome_email = YesNoSwitch(
        label=_("Send a welcoming e-mail to users on their first sign-ons"),
        required=False,
    )

    oauth2_json_id_path = forms.CharField(
        label=_("User ID path"),
        max_length=200,
        required=False,
    )
    oauth2_json_name_path = forms.CharField(
        label=_("User name path"),
        max_length=200,
        required=False,
    )
    oauth2_json_email_path = forms.CharField(
        label=_("User e-mail path"),
        max_length=200,
        required=False,
    )
    oauth2_json_avatar_path = forms.CharField(
        label=_("User avatar URL path"),
        help_text=_("Optional, leave empty to don't download avatar from provider."),
        max_length=200,
        required=False,
    )

    def clean_oauth2_scopes(self):
        # Remove duplicates and extra spaces, keep order of scopes
        clean_scopes = []
        for scope in self.cleaned_data["oauth2_scopes"].split():
            scope = scope.strip()
            if scope and scope not in clean_scopes:
                clean_scopes.append(scope)

        return " ".join(clean_scopes) or None

    def clean_oauth2_token_extra_headers(self):
        return clean_headers(self.cleaned_data["oauth2_token_extra_headers"])

    def clean_oauth2_user_extra_headers(self):
        return clean_headers(self.cleaned_data["oauth2_user_extra_headers"])

    def clean(self):
        data = super().clean()

        if not data.get("enable_oauth2_client"):
            return data

        required_data = [data[key] for key in data if key not in OAUTH2_OPTIONAL_FIELDS]

        if not all(required_data):
            data["enable_oauth2_client"] = False

            messages.error(
                self.request,
                gettext(
                    "You need to complete the configuration before you will be able to "
                    "enable OAuth 2 on your site."
                ),
            )

        return data


def clean_headers(headers_value):
    clean_headers = {}
    for header in headers_value.splitlines():
        header = header.strip()
        if not header:
            continue
        if ":" not in header:
            raise forms.ValidationError(
                gettext(
                    '"%(header)s" is not a valid header. '
                    'It\'s missing a colon (":").'
                )
                % {"header": header},
            )

        name, value = [part.strip() for part in header.split(":", 1)]

        if not name:
            raise forms.ValidationError(
                gettext(
                    '"%(header)s" is not a valid header. '
                    'It\'s missing a header name before the colon (":").'
                )
                % {"header": header},
            )

        if name in clean_headers:
            raise forms.ValidationError(
                gettext('"%(header)s" header is entered more than once.')
                % {"header": name},
            )

        if not value:
            raise forms.ValidationError(
                gettext(
                    '"%(header)s" is not a valid header. '
                    'It\'s missing a header value after the colon (":").'
                )
                % {"header": header},
            )

        clean_headers[name] = value

    return "\n".join([f"{key}: {value}" for key, value in clean_headers.items()])



import time


def show_progress(command, step, total, since=None):
    progress = step * 100 // total
    filled = progress // 2
    blank = 50 - filled

    template = "\r%(step)s (%(progress)s%%) [%(progressbar)s]%(estimation)s"
    variables = {
        "step": str(step).rjust(len(str(total))),
        "progress": str(progress).rjust(3),
        "progressbar": "".join(["=" * filled, " " * blank]),
        "estimation": get_estimation_str(since, progress, step, total),
    }

    command.stdout.write(template % variables, ending="")
    command.stdout.flush()


def get_estimation_str(since, progress, step, total):
    if not since:
        return ""

    progress_float = float(step) * 100.0 / float(total)
    if progress_float == 0:
        return " --:--:-- est."

    step_time = (time.time() - since) / progress_float
    estimated_time = (100 - progress) * step_time
    clock = time.strftime("%H:%M:%S", time.gmtime(estimated_time))
    return " %s est." % clock



import markdown
from markdown.extensions.fenced_code import FencedCodeExtension

from .bbcode.code import CodeBlockExtension
from .bbcode.hr import BBCodeHRProcessor
from .bbcode.inline import bold, image, italics, underline, url
from .bbcode.quote import QuoteExtension
from .bbcode.spoiler import SpoilerExtension
from .htmlparser import parse_html_string, print_html_string
from .links import clean_links, linkify_texts
from .md.shortimgs import ShortImagesExtension
from .md.strikethrough import StrikethroughExtension
from .mentions import add_mentions
from .pipeline import pipeline


def parse(
    text,
    request,
    poster,
    allow_mentions=True,
    allow_links=True,
    allow_images=True,
    allow_blocks=True,
    force_shva=False,
):
    """
    Message parser

    Utility for flavours to call

    Breaks text into paragraphs, supports code, spoiler and quote blocks,
    headers, lists, images, spoilers, text styles

    Returns dict object
    """
    md = md_factory(
        allow_links=allow_links, allow_images=allow_images, allow_blocks=allow_blocks
    )

    parsing_result = {
        "original_text": text,
        "parsed_text": "",
        "markdown": md,
        "mentions": [],
        "images": [],
        "internal_links": [],
        "outgoing_links": [],
    }

    # Parse text
    parsed_text = md.convert(text)

    # Clean and store parsed text
    parsing_result["parsed_text"] = parsed_text.strip()

    # Run additional operations
    if allow_mentions or allow_links or allow_images:
        root_node = parse_html_string(parsing_result["parsed_text"])

        if allow_links:
            linkify_texts(root_node)

        if allow_mentions:
            add_mentions(parsing_result, root_node)

        if allow_links or allow_images:
            clean_links(request, parsing_result, root_node, force_shva)

        parsing_result["parsed_text"] = print_html_string(root_node)

    # Let plugins do their magic
    parsing_result = pipeline.process_result(parsing_result)

    return parsing_result


def md_factory(allow_links=True, allow_images=True, allow_blocks=True):
    """creates and configures markdown object"""
    md = markdown.Markdown(extensions=["markdown.extensions.nl2br"])

    # Remove HTML allowances
    md.preprocessors.deregister("html_block")
    md.inlinePatterns.deregister("html")

    # Remove references
    md.parser.blockprocessors.deregister("reference")
    md.inlinePatterns.deregister("reference")
    md.inlinePatterns.deregister("image_reference")
    md.inlinePatterns.deregister("short_reference")

    # Add [b], [i], [u]
    md.inlinePatterns.register(bold, "bb_b", 55)
    md.inlinePatterns.register(italics, "bb_i", 55)
    md.inlinePatterns.register(underline, "bb_u", 55)

    # Add ~~deleted~~
    strikethrough_md = StrikethroughExtension()
    strikethrough_md.extendMarkdown(md)

    if allow_links:
        # Add [url]
        md.inlinePatterns.register(url(md), "bb_url", 155)
    else:
        # Remove links
        md.inlinePatterns.deregister("link")
        md.inlinePatterns.deregister("autolink")
        md.inlinePatterns.deregister("automail")

    if allow_images:
        # Add [img]
        md.inlinePatterns.register(image(md), "bb_img", 145)
        short_images_md = ShortImagesExtension()
        short_images_md.extendMarkdown(md)
    else:
        # Remove images
        md.inlinePatterns.deregister("image_link")

    if allow_blocks:
        # Add [hr] and [quote] blocks
        md.parser.blockprocessors.register(BBCodeHRProcessor(md.parser), "bb_hr", 45)

        fenced_code = FencedCodeExtension(lang_prefix="language-")
        fenced_code.extendMarkdown(md)

        code_bbcode = CodeBlockExtension()
        code_bbcode.extendMarkdown(md)

        quote_bbcode = QuoteExtension()
        quote_bbcode.extendMarkdown(md)

        spoiler_bbcode = SpoilerExtension()
        spoiler_bbcode.extendMarkdown(md)
    else:
        # Remove blocks
        md.parser.blockprocessors.deregister("hashheader")
        md.parser.blockprocessors.deregister("setextheader")
        md.parser.blockprocessors.deregister("code")
        md.parser.blockprocessors.deregister("quote")
        md.parser.blockprocessors.deregister("hr")
        md.parser.blockprocessors.deregister("olist")
        md.parser.blockprocessors.deregister("ulist")

    return pipeline.extend_markdown(md)



from django.http import Http404
from rest_framework.decorators import api_view
from rest_framework.response import Response


@api_view()
def question(request):
    if not request.settings.qa_question:
        raise Http404()

    return Response(
        {
            "question": request.settings.qa_question,
            "help_text": request.settings.qa_help_text,
        }
    )



from django.test import TestCase

from ..models import Ban


class BansManagerTests(TestCase):
    def setUp(self):
        Ban.objects.bulk_create(
            [
                Ban(check_type=Ban.USERNAME, banned_value="user"),
                Ban(check_type=Ban.EMAIL, banned_value="user@example.com"),
                Ban(check_type=Ban.IP, banned_value="127.0.0.1"),
            ]
        )

    def test_get_ban_for_banned_name(self):
        """get_ban finds ban for given username"""
        self.assertIsNotNone(Ban.objects.get_ban(username="User"))
        with self.assertRaises(Ban.DoesNotExist):
            Ban.objects.get_ban(username="OtherUser")

    def test_get_ban_for_banned_email(self):
        """get_ban finds ban for given email"""
        self.assertIsNotNone(Ban.objects.get_ban(email="user@example.com"))
        with self.assertRaises(Ban.DoesNotExist):
            Ban.objects.get_ban(email="otheruser@example.com")

    def test_get_ban_for_banned_ip(self):
        """get_ban finds ban for given ip"""
        self.assertIsNotNone(Ban.objects.get_ban(ip="127.0.0.1"))
        with self.assertRaises(Ban.DoesNotExist):
            Ban.objects.get_ban(ip="42.0.0.1")

    def test_get_ban_for_all_values(self):
        """get_ban finds ban for given values"""
        valid_kwargs = {"username": "User", "ip": "42.51.52.51"}
        self.assertIsNotNone(Ban.objects.get_ban(**valid_kwargs))

        invalid_kwargs = {"username": "OtherUser", "ip": "42.51.52.51"}
        with self.assertRaises(Ban.DoesNotExist):
            Ban.objects.get_ban(**invalid_kwargs)


class BanTests(TestCase):
    def test_check_value_literal(self):
        """ban correctly tests given values"""
        test_ban = Ban(banned_value="user")

        self.assertTrue(test_ban.check_value("User"))
        self.assertFalse(test_ban.check_value("OtherUser"))
        self.assertFalse(test_ban.check_value("UserOther"))

    def test_check_value_starts_with(self):
        """ban correctly tests given values"""
        test_ban = Ban(banned_value="user*")

        self.assertTrue(test_ban.check_value("User"))
        self.assertTrue(test_ban.check_value("UserOther"))
        self.assertFalse(test_ban.check_value("NewUser"))

    def test_check_value_middle_match(self):
        """ban correctly tests given values"""
        test_ban = Ban(banned_value="u*r")

        self.assertTrue(test_ban.check_value("User"))
        self.assertTrue(test_ban.check_value("UserOther"))
        self.assertFalse(test_ban.check_value("NewUser"))
        self.assertFalse(test_ban.check_value("UserNew"))

    def test_check_value_ends_witch(self):
        """ban correctly tests given values"""
        test_ban = Ban(banned_value="*user")

        self.assertTrue(test_ban.check_value("User"))
        self.assertTrue(test_ban.check_value("OtherUser"))
        self.assertFalse(test_ban.check_value("UserOther"))



from django.urls import include, path


class URLPatterns:
    def __init__(self):
        self.namespace_patterns = None
        self._namespaces = []
        self._patterns = []
        self._urlpatterns = None

    def namespace(self, url_path, namespace, parent=None):
        self._namespaces.append(
            {"path": url_path, "parent": parent, "namespace": namespace}
        )

    def patterns(self, namespace, *new_patterns):
        self._patterns.append({"namespace": namespace, "urlpatterns": new_patterns})

    def single_pattern(self, url_path, namespace, parent, view):
        self.namespace(url_path, namespace, parent)
        self.patterns(":".join((parent, namespace)), path("", view, name="index"))

    def get_child_patterns(self, parent):
        prefix = "%s:" % parent if parent else ""

        namespace_urlpatterns = self.namespace_patterns.get(parent, [])
        for namespace in self._namespaces:
            if namespace["parent"] == parent:
                prefixed_namespace = prefix + namespace["namespace"]
                child_patterns = self.get_child_patterns(prefixed_namespace)
                included_patterns = include(
                    (child_patterns, namespace["namespace"]),
                    namespace=namespace["namespace"],
                )
                namespace_urlpatterns.append(path(namespace["path"], included_patterns))
        return namespace_urlpatterns

    def sum_registered_patters(self):
        all_patterns = {}
        for urls in self._patterns:
            namespace = urls["namespace"]
            added_patterns = urls["urlpatterns"]
            all_patterns.setdefault(namespace, []).extend(added_patterns)

        self.namespace_patterns = all_patterns

    def build_root_urlpatterns(self):
        root_urlpatterns = []
        for namespace in self._namespaces:
            if not namespace["parent"]:
                child_patterns = self.get_child_patterns(namespace["namespace"])
                included_patterns = include(
                    (child_patterns, namespace["namespace"]),
                    namespace=namespace["namespace"],
                )
                root_urlpatterns.append(path(namespace["path"], included_patterns))

        return root_urlpatterns

    def build_urlpatterns(self):
        self.sum_registered_patters()
        return self.build_root_urlpatterns()

    def __call__(self):
        if self._urlpatterns:
            return self._urlpatterns

        self._urlpatterns = self.build_urlpatterns()
        self._namespaces = []
        self._patterns = []
        return self._urlpatterns


urlpatterns = URLPatterns()



from datetime import timedelta

from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from ...acl import algebra
from ...acl.decorators import return_boolean
from ...acl.models import Role

__all__ = [
    "allow_delete_user",
    "can_delete_user",
    "allow_delete_own_account",
    "can_delete_own_account",
]


class PermissionsForm(forms.Form):
    legend = _("Deleting users")

    can_delete_users_newer_than = forms.IntegerField(
        label=_("Maximum age of deleted account (in days)"),
        help_text=_("Enter zero to disable this check."),
        min_value=0,
        initial=0,
    )
    can_delete_users_with_less_posts_than = forms.IntegerField(
        label=_("Maximum number of posts on deleted account"),
        help_text=_("Enter zero to disable this check."),
        min_value=0,
        initial=0,
    )


def change_permissions_form(role):
    if isinstance(role, Role) and role.special_role != "anonymous":
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {
        "can_delete_users_newer_than": 0,
        "can_delete_users_with_less_posts_than": 0,
    }
    new_acl.update(acl)

    return algebra.sum_acls(
        new_acl,
        roles=roles,
        key=key_name,
        can_delete_users_newer_than=algebra.greater,
        can_delete_users_with_less_posts_than=algebra.greater,
    )


def add_acl_to_user(user_acl, target):
    target.acl["can_delete"] = can_delete_user(user_acl, target)
    if target.acl["can_delete"]:
        target.acl["can_moderate"] = True


def register_with(registry):
    registry.acl_annotator(get_user_model(), add_acl_to_user)


def allow_delete_user(user_acl, target):
    newer_than = user_acl["can_delete_users_newer_than"]
    less_posts_than = user_acl["can_delete_users_with_less_posts_than"]
    if not newer_than and not less_posts_than:
        raise PermissionDenied(_("You can't delete users."))

    if user_acl["user_id"] == target.id:
        raise PermissionDenied(_("You can't delete your account."))
    if target.is_staff or target.is_superuser:
        raise PermissionDenied(_("You can't delete administrators."))

    if newer_than:
        if target.joined_on < timezone.now() - timedelta(days=newer_than):
            message = ngettext(
                "You can't delete users that are members for more than %(days)s day.",
                "You can't delete users that are members for more than %(days)s days.",
                newer_than,
            )
            raise PermissionDenied(message % {"days": newer_than})
    if less_posts_than:
        if target.posts > less_posts_than:
            message = ngettext(
                "You can't delete users that made more than %(posts)s post.",
                "You can't delete users that made more than %(posts)s posts.",
                less_posts_than,
            )
            raise PermissionDenied(message % {"posts": less_posts_than})


can_delete_user = return_boolean(allow_delete_user)


def allow_delete_own_account(settings, user, target):
    if user.id != target.id:
        raise PermissionDenied(_("You can't delete other users accounts."))
    if not settings.allow_delete_own_account and not user.is_deleting_account:
        raise PermissionDenied(_("You can't delete your account."))
    if user.is_staff or user.is_superuser:
        raise PermissionDenied(
            _("You can't delete your account because you are an administrator.")
        )


can_delete_own_account = return_boolean(allow_delete_own_account)



from django.db import migrations


def delete_oauth2_token_method_setting(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    Setting.objects.filter(setting="oauth2_token_method").delete()


class Migration(migrations.Migration):

    dependencies = [
        ("misago_conf", "0008_delete_sso_settings"),
    ]

    operations = [migrations.RunPython(delete_oauth2_token_method_setting)]



from ..context_processors import hooks


def test_context_processors_hook_can_inject_context(mocker):
    plugin = mocker.Mock(return_value={"plugin": True})
    mocker.patch("misago.core.context_processors.context_processors_hooks", [plugin])
    context = hooks(None)
    assert context == {"plugin": True}



from django import forms
from django.utils.translation import gettext_lazy as _

from .base import OAuthProviderForm


class FacebookForm(OAuthProviderForm):
    key = forms.CharField(label=_("App ID"), required=False)
    secret = forms.CharField(label=_("App Secret"), required=False)



from celery import shared_task
from django.contrib.auth import get_user_model

from ...conf.shortcuts import get_dynamic_settings
from ..deletesrecord import record_user_deleted_by_staff

User = get_user_model()


@shared_task
def delete_user_with_content(pk):
    try:
        user = User.objects.get(pk=pk, is_staff=False, is_superuser=False)
    except User.DoesNotExist:
        pass
    else:
        settings = get_dynamic_settings()
        user.delete(anonymous_username=settings.anonymous_username, delete_content=True)
        record_user_deleted_by_staff()



from django.urls import reverse

from .. import test
from ..test import patch_category_acl
from .test_threads_api import ThreadsApiTestCase


class ThreadPostEditsApiTestCase(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.post = test.reply_thread(self.thread, poster=self.user)

        self.api_link = reverse(
            "misago:api:thread-post-edits",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    def mock_edit_record(self):
        edits_record = [
            self.post.edits_record.create(
                category=self.category,
                thread=self.thread,
                editor=self.user,
                editor_name=self.user.username,
                editor_slug=self.user.slug,
                edited_from="Original body",
                edited_to="First Edit",
            ),
            self.post.edits_record.create(
                category=self.category,
                thread=self.thread,
                editor_name="Deleted",
                editor_slug="deleted",
                edited_from="First Edit",
                edited_to="Second Edit",
            ),
            self.post.edits_record.create(
                category=self.category,
                thread=self.thread,
                editor=self.user,
                editor_name=self.user.username,
                editor_slug=self.user.slug,
                edited_from="Second Edit",
                edited_to="Last Edit",
            ),
        ]

        self.post.original = "Last Edit"
        self.post.parsed = "<p>Last Edit</p>"
        self.post.save()

        return edits_record


class ThreadPostGetEditTests(ThreadPostEditsApiTestCase):
    def test_no_edits(self):
        """api returns 403 if post has no edits record"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "Edits record is unavailable for this post."}
        )

        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "Edits record is unavailable for this post."}
        )

    def test_empty_edit_id(self):
        """api handles empty edit in querystring"""
        response = self.client.get("%s?edit=" % self.api_link)
        self.assertEqual(response.status_code, 404)

    def test_invalid_edit_id(self):
        """api handles invalid edit in querystring"""
        response = self.client.get("%s?edit=dsa67d8sa68" % self.api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_edit_id(self):
        """api handles nonexistant edit in querystring"""
        response = self.client.get("%s?edit=1321" % self.api_link)
        self.assertEqual(response.status_code, 404)

    def test_get_last_edit(self):
        """api returns last edit record"""
        edits = self.mock_edit_record()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["id"], edits[-1].id)
        self.assertIsNone(response_json["next"])
        self.assertEqual(response_json["previous"], edits[1].id)

    def test_get_middle_edit(self):
        """api returns middle edit record"""
        edits = self.mock_edit_record()

        response = self.client.get("%s?edit=%s" % (self.api_link, edits[1].id))
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["id"], edits[1].id)
        self.assertEqual(response_json["next"], edits[-1].id)
        self.assertEqual(response_json["previous"], edits[0].id)

    def test_get_first_edit(self):
        """api returns middle edit record"""
        edits = self.mock_edit_record()

        response = self.client.get("%s?edit=%s" % (self.api_link, edits[0].id))
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["id"], edits[0].id)
        self.assertEqual(response_json["next"], edits[1].id)
        self.assertIsNone(response_json["previous"])


class ThreadPostPostEditTests(ThreadPostEditsApiTestCase):
    def setUp(self):
        super().setUp()
        self.edits = self.mock_edit_record()

    @patch_category_acl({"can_edit_posts": 2})
    def test_empty_edit_id(self):
        """api handles empty edit in querystring"""
        response = self.client.post("%s?edit=" % self.api_link)
        self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_edit_posts": 2})
    def test_invalid_edit_id(self):
        """api handles invalid edit in querystring"""
        response = self.client.post("%s?edit=dsa67d8sa68" % self.api_link)
        self.assertEqual(response.status_code, 404)

    @patch_category_acl({"can_edit_posts": 2})
    def test_nonexistant_edit_id(self):
        """api handles nonexistant edit in querystring"""
        response = self.client.post("%s?edit=1321" % self.api_link)
        self.assertEqual(response.status_code, 404)

    def test_anonymous(self):
        """only signed in users can rever ports"""
        self.logout_user()

        response = self.client.post("%s?edit=%s" % (self.api_link, self.edits[0].id))
        self.assertEqual(response.status_code, 403)

    @patch_category_acl({"can_edit_posts": 0})
    def test_no_permission(self):
        """api validates permission to revert post"""
        response = self.client.post("%s?edit=1321" % self.api_link)
        self.assertEqual(response.status_code, 403)

    @patch_category_acl({"can_edit_posts": 2})
    def test_revert_post(self):
        """api reverts post to version from before specified edit"""
        response = self.client.post("%s?edit=%s" % (self.api_link, self.edits[0].id))
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["edits"], 1)
        self.assertEqual(response_json["content"], "<p>Original body</p>")

        self.assertEqual(self.post.edits_record.count(), 4)

        edit = self.post.edits_record.first()
        self.assertEqual(edit.edited_from, self.post.original)
        self.assertEqual(edit.edited_to, "Original body")



from django.utils import timezone

from ..threads.models import Thread
from .englishcorpus import EnglishCorpus
from .posts import get_fake_hidden_post, get_fake_post, get_fake_unapproved_post

corpus_short = EnglishCorpus(max_length=150)


def get_fake_thread(fake, category, starter=None):
    thread = _create_base_thread(fake, category)
    thread.first_post = get_fake_post(fake, thread, starter)
    thread.save(update_fields=["first_post"])
    return thread


def get_fake_closed_thread(fake, category, starter=None):
    thread = get_fake_thread(fake, category)
    thread.is_closed = True
    thread.save(update_fields=["is_closed"])
    return thread


def get_fake_hidden_thread(fake, category, starter=None, hidden_by=None):
    thread = _create_base_thread(fake, category)
    thread.first_post = get_fake_hidden_post(fake, thread, starter, hidden_by)
    thread.is_hidden = True
    thread.save(update_fields=["first_post", "is_hidden"])
    return thread


def get_fake_unapproved_thread(fake, category, starter=None):
    thread = _create_base_thread(fake, category)
    thread.first_post = get_fake_unapproved_post(fake, thread, starter)
    thread.is_unapproved = True
    thread.save(update_fields=["first_post", "is_unapproved"])
    return thread


def _create_base_thread(fake, category):
    started_on = timezone.now()
    thread = Thread(
        category=category,
        started_on=started_on,
        starter_name="-",
        starter_slug="-",
        last_post_on=started_on,
        last_poster_name="-",
        last_poster_slug="-",
        replies=0,
    )

    # Sometimes thread ends with slug being set to empty string
    while not thread.slug:
        thread.set_title(corpus_short.random_sentence())

    thread.save()

    return thread



from datetime import timedelta

from django.urls import reverse
from django.utils import timezone

from ...acl.test import patch_user_acl
from ..serializers.poll import MAX_POLL_OPTIONS
from ..test import patch_category_acl
from .test_thread_poll_api import ThreadPollApiTestCase


class ThreadPollEditTests(ThreadPollApiTestCase):
    def setUp(self):
        super().setUp()

        self.mock_poll()

    def test_anonymous(self):
        """api requires you to sign in to edit poll"""
        self.logout_user()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_invalid_thread_id(self):
        """api validates that thread id is integer"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": "kjha6dsa687sa", "pk": self.poll.pk},
        )

        response = self.put(api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_thread_id(self):
        """api validates that thread exists"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": self.thread.pk + 1, "pk": self.poll.pk},
        )

        response = self.put(api_link)
        self.assertEqual(response.status_code, 404)

    def test_invalid_poll_id(self):
        """api validates that poll id is integer"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": "sad98as7d97sa98"},
        )

        response = self.put(api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_poll_id(self):
        """api validates that poll exists"""
        api_link = reverse(
            "misago:api:thread-poll-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.poll.pk + 123},
        )

        response = self.put(api_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_edit_polls": 0})
    def test_no_permission(self):
        """api validates that user has permission to edit poll in thread"""
        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't edit polls."})

    @patch_user_acl({"can_edit_polls": 1, "poll_edit_time": 5})
    def test_no_permission_timeout(self):
        """api validates that user's window to edit poll in thread has closed"""
        self.poll.posted_on = timezone.now() - timedelta(minutes=15)
        self.poll.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't edit polls that are older than 5 minutes."},
        )

    @patch_user_acl({"can_edit_polls": 1})
    def test_no_permission_poll_closed(self):
        """api validates that user's window to edit poll in thread has closed"""
        self.poll.posted_on = timezone.now() - timedelta(days=15)
        self.poll.length = 5
        self.poll.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This poll is over. You can't edit it."}
        )

    @patch_user_acl({"can_edit_polls": 1})
    def test_no_permission_other_user_poll(self):
        """api validates that user has permission to edit other user poll in thread"""
        self.poll.poster = None
        self.poll.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't edit other users polls in this category."},
        )

    @patch_user_acl({"can_edit_polls": 1})
    @patch_category_acl({"can_close_threads": False})
    def test_no_permission_closed_thread(self):
        """api validates that user has permission to edit poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't edit polls in it."},
        )

    @patch_user_acl({"can_edit_polls": 1})
    @patch_category_acl({"can_close_threads": True})
    def test_closed_thread(self):
        """api validates that user has permission to edit poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_user_acl({"can_edit_polls": 1})
    @patch_category_acl({"can_close_threads": False})
    def test_no_permission_closed_category(self):
        """api validates that user has permission to edit poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't edit polls in it."},
        )

    @patch_user_acl({"can_edit_polls": 1})
    @patch_category_acl({"can_close_threads": True})
    def test_closed_category(self):
        """api validates that user has permission to edit poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 400)

    def test_empty_data(self):
        """api handles empty request data"""
        response = self.put(self.api_link)
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(len(response_json), 4)

    def test_length_validation(self):
        """api validates poll's length"""
        response = self.put(self.api_link, data={"length": -1})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["length"],
            ["Ensure this value is greater than or equal to 0."],
        )

        response = self.put(self.api_link, data={"length": 200})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["length"], ["Ensure this value is less than or equal to 180."]
        )

    def test_question_validation(self):
        """api validates question length"""
        response = self.put(self.api_link, data={"question": "abcd" * 255})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["question"],
            ["Ensure this field has no more than 255 characters."],
        )

    def test_validate_choice_length(self):
        """api validates single choice length"""
        response = self.put(
            self.api_link, data={"choices": [{"hash": "qwertyuiopas", "label": ""}]}
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"], ["One or more poll choices are invalid."]
        )

        response = self.put(
            self.api_link,
            data={"choices": [{"hash": "qwertyuiopas", "label": "abcd" * 255}]},
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"], ["One or more poll choices are invalid."]
        )

    def test_validate_two_choices(self):
        """api validates that there are at least two choices in poll"""
        response = self.put(self.api_link, data={"choices": [{"label": "Choice"}]})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"],
            ["You need to add at least two choices to a poll."],
        )

    def test_validate_max_choices(self):
        """api validates that there are no more choices in poll than allowed number"""
        response = self.put(
            self.api_link,
            data={"choices": [{"label": "Choice"}] * (MAX_POLL_OPTIONS + 1)},
        )
        self.assertEqual(response.status_code, 400)

        error_formats = (MAX_POLL_OPTIONS, MAX_POLL_OPTIONS + 1)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"],
            [
                "You can't add more than %s options to a single poll (added %s)."
                % error_formats
            ],
        )

    def test_allowed_choices_validation(self):
        """api validates allowed choices number"""
        response = self.put(self.api_link, data={"allowed_choices": 0})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["allowed_choices"],
            ["Ensure this value is greater than or equal to 1."],
        )

        response = self.put(
            self.api_link,
            data={
                "length": 0,
                "question": "Lorem ipsum",
                "allowed_choices": 3,
                "choices": [{"label": "Choice"}, {"label": "Choice"}],
            },
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["non_field_errors"],
            ["Number of allowed choices can't be greater than number of all choices."],
        )

    def test_poll_all_choices_replaced(self):
        """api edits all poll choices out"""
        response = self.put(
            self.api_link,
            data={
                "length": 40,
                "question": "Select two best colors",
                "allowed_choices": 2,
                "allow_revotes": True,
                "is_public": True,
                "choices": [
                    {"label": "\nRed  "},
                    {"label": "Green"},
                    {"label": "Blue"},
                ],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["poster_name"], self.user.username)
        self.assertEqual(response_json["length"], 40)
        self.assertEqual(response_json["question"], "Select two best colors")
        self.assertEqual(response_json["allowed_choices"], 2)
        self.assertTrue(response_json["allow_revotes"])

        # you can't change poll's type after its posted
        self.assertFalse(response_json["is_public"])

        # choices were updated
        self.assertEqual(len(response_json["choices"]), 3)
        self.assertEqual(len({c["hash"] for c in response_json["choices"]}), 3)
        self.assertEqual(
            [c["label"] for c in response_json["choices"]], ["Red", "Green", "Blue"]
        )
        self.assertEqual([c["votes"] for c in response_json["choices"]], [0, 0, 0])
        self.assertEqual(
            [c["selected"] for c in response_json["choices"]], [False, False, False]
        )

        # votes were removed
        self.assertEqual(response_json["votes"], 0)
        self.assertEqual(self.poll.pollvote_set.count(), 0)

        self.assertEqual(self.user.audittrail_set.count(), 1)

    def test_poll_current_choices_edited(self):
        """api edits current poll choices"""
        response = self.put(
            self.api_link,
            data={
                "length": 40,
                "question": "Select two best colors",
                "allowed_choices": 2,
                "allow_revotes": True,
                "is_public": True,
                "choices": [
                    {"hash": "aaaaaaaaaaaa", "label": "\nFirst  ", "votes": 5555},
                    {"hash": "bbbbbbbbbbbb", "label": "Second", "votes": 5555},
                    {"hash": "gggggggggggg", "label": "Third", "votes": 5555},
                    {"hash": "dddddddddddd", "label": "Fourth", "votes": 5555},
                ],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["poster_name"], self.user.username)
        self.assertEqual(response_json["length"], 40)
        self.assertEqual(response_json["question"], "Select two best colors")
        self.assertEqual(response_json["allowed_choices"], 2)
        self.assertTrue(response_json["allow_revotes"])

        # you can't change poll's type after its posted
        self.assertFalse(response_json["is_public"])

        # choices were updated
        self.assertEqual(len(response_json["choices"]), 4)
        self.assertEqual(
            response_json["choices"],
            [
                {
                    "hash": "aaaaaaaaaaaa",
                    "label": "First",
                    "votes": 1,
                    "selected": False,
                },
                {
                    "hash": "bbbbbbbbbbbb",
                    "label": "Second",
                    "votes": 0,
                    "selected": False,
                },
                {
                    "hash": "gggggggggggg",
                    "label": "Third",
                    "votes": 2,
                    "selected": True,
                },
                {
                    "hash": "dddddddddddd",
                    "label": "Fourth",
                    "votes": 1,
                    "selected": True,
                },
            ],
        )

        # no votes were removed
        self.assertEqual(response_json["votes"], 4)
        self.assertEqual(self.poll.pollvote_set.count(), 4)

        self.assertEqual(self.user.audittrail_set.count(), 1)

    def test_poll_some_choices_edited(self):
        """api edits some poll choices"""
        response = self.put(
            self.api_link,
            data={
                "length": 40,
                "question": "Select two best colors",
                "allowed_choices": 2,
                "allow_revotes": True,
                "is_public": True,
                "choices": [
                    {"hash": "aaaaaaaaaaaa", "label": "\nFirst ", "votes": 5555},
                    {"hash": "bbbbbbbbbbbb", "label": "Second", "votes": 5555},
                    {"hash": "dsadsadsa788", "label": "New Option", "votes": 5555},
                ],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["poster_name"], self.user.username)
        self.assertEqual(response_json["length"], 40)
        self.assertEqual(response_json["question"], "Select two best colors")
        self.assertEqual(response_json["allowed_choices"], 2)
        self.assertTrue(response_json["allow_revotes"])

        # you can't change poll's type after its posted
        self.assertFalse(response_json["is_public"])

        # choices were updated
        self.assertEqual(len(response_json["choices"]), 3)
        self.assertEqual(
            response_json["choices"],
            [
                {
                    "hash": "aaaaaaaaaaaa",
                    "label": "First",
                    "votes": 1,
                    "selected": False,
                },
                {
                    "hash": "bbbbbbbbbbbb",
                    "label": "Second",
                    "votes": 0,
                    "selected": False,
                },
                {
                    "hash": response_json["choices"][2]["hash"],
                    "label": "New Option",
                    "votes": 0,
                    "selected": False,
                },
            ],
        )

        # no votes were removed
        self.assertEqual(response_json["votes"], 1)
        self.assertEqual(self.poll.pollvote_set.count(), 1)

        self.assertEqual(self.user.audittrail_set.count(), 1)

    @patch_user_acl({"can_edit_polls": 2, "poll_edit_time": 5})
    def test_moderate_user_poll(self):
        """api edits all poll choices out in other users poll, even if its over"""
        self.poll.poster = None
        self.poll.posted_on = timezone.now() - timedelta(days=15)
        self.poll.length = 5
        self.poll.save()

        response = self.put(
            self.api_link,
            data={
                "length": 40,
                "question": "Select two best colors",
                "allowed_choices": 2,
                "allow_revotes": True,
                "is_public": True,
                "choices": [
                    {"label": "\nRed  "},
                    {"label": "Green"},
                    {"label": "Blue"},
                ],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["poster_name"], self.user.username)
        self.assertEqual(response_json["length"], 40)
        self.assertEqual(response_json["question"], "Select two best colors")
        self.assertEqual(response_json["allowed_choices"], 2)
        self.assertTrue(response_json["allow_revotes"])

        # you can't change poll's type after its posted
        self.assertFalse(response_json["is_public"])

        # choices were updated
        self.assertEqual(len(response_json["choices"]), 3)
        self.assertEqual(len({c["hash"] for c in response_json["choices"]}), 3)
        self.assertEqual(
            [c["label"] for c in response_json["choices"]], ["Red", "Green", "Blue"]
        )
        self.assertEqual([c["votes"] for c in response_json["choices"]], [0, 0, 0])
        self.assertEqual(
            [c["selected"] for c in response_json["choices"]], [False, False, False]
        )

        # votes were removed
        self.assertEqual(response_json["votes"], 0)
        self.assertEqual(self.poll.pollvote_set.count(), 0)

        self.assertEqual(self.user.audittrail_set.count(), 1)



from unittest.mock import patch

from ..validation import filter_user_data


def test_new_user_data_is_filtered_using_default_filters(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": "New User",
            "email": "oauth2@example.com",
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "New_User",
        "email": "oauth2@example.com",
        "avatar": None,
    }


def test_existing_user_data_is_filtered_using_default_filters(user, request):
    filtered_data = filter_user_data(
        request,
        user,
        {
            "id": "242132",
            "name": user.username,
            "email": user.email,
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": user.username,
        "email": user.email,
        "avatar": None,
    }


def test_empty_user_name_is_replaced_with_placeholder_one(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": None,
            "email": "oauth2@example.com",
            "avatar": None,
        },
    )

    assert len(filtered_data["name"].strip())


def test_missing_user_name_is_replaced_with_placeholder_one(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": None,
            "email": "oauth2@example.com",
            "avatar": None,
        },
    )

    assert len(filtered_data["name"].strip())


def test_user_name_is_converted_to_str(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": 123456,
            "email": "oauth2@example.com",
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "123456",
        "email": "oauth2@example.com",
        "avatar": None,
    }


def test_user_email_is_converted_to_str(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": "New User",
            "email": [1, 2, 4, "d"],
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "New_User",
        "email": "[1, 2, 4, 'd']",
        "avatar": None,
    }


def test_missing_user_email_is_set_as_empty_str(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": "New User",
            "email": None,
            "avatar": None,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "New_User",
        "email": "",
        "avatar": None,
    }


def test_user_avatar_is_converted_to_str(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": "New User",
            "email": "oauth2@example.com",
            "avatar": 123456,
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "New_User",
        "email": "oauth2@example.com",
        "avatar": "123456",
    }


def test_empty_user_avatar_is_filtered_to_none(db, request):
    filtered_data = filter_user_data(
        request,
        None,
        {
            "id": "242132",
            "name": "New User",
            "email": "oauth2@example.com",
            "avatar": "",
        },
    )

    assert filtered_data == {
        "id": "242132",
        "name": "New_User",
        "email": "oauth2@example.com",
        "avatar": None,
    }


def user_request_filter(request, user, user_data):
    assert request


def user_id_filter(request, user, user_data):
    return {
        "id": "".join(reversed(user_data["id"])),
        "name": user_data["name"],
        "email": user_data["email"],
        "avatar": user_data["avatar"],
    }


def user_name_filter(request, user, user_data):
    return {
        "id": user_data["id"],
        "name": "".join(reversed(user_data["name"])),
        "email": user_data["email"],
        "avatar": user_data["avatar"],
    }


def user_email_filter(request, user, user_data):
    return {
        "id": user_data["id"],
        "name": user_data["name"],
        "email": "filtered_%s" % user_data["email"],
        "avatar": user_data["avatar"],
    }


def test_new_user_data_is_filtered_using_custom_filters(db, request):
    def user_is_none_filter(request, user, user_data):
        assert user is None

    with patch(
        "misago.oauth2.validation.oauth2_user_data_filters",
        [
            user_request_filter,
            user_is_none_filter,
            user_id_filter,
            user_name_filter,
            user_email_filter,
        ],
    ):
        filtered_data = filter_user_data(
            request,
            None,
            {
                "id": "1234",
                "name": "New User",
                "email": "oauth2@example.com",
                "avatar": None,
            },
        )

        assert filtered_data == {
            "id": "4321",
            "name": "resU weN",
            "email": "filtered_oauth2@example.com",
            "avatar": None,
        }


def test_existing_user_data_is_filtered_using_custom_filters(user, request):
    def user_is_set_filter(request, user_obj, user_data):
        assert user_obj == user
        return {
            "id": str(user_obj.id),
            "name": user_data["name"],
            "email": user_data["email"],
            "avatar": user_data["avatar"],
        }

    with patch(
        "misago.oauth2.validation.oauth2_user_data_filters",
        [
            user_request_filter,
            user_is_set_filter,
            user_id_filter,
            user_name_filter,
            user_email_filter,
        ],
    ):
        filtered_data = filter_user_data(
            request,
            user,
            {
                "id": "1234",
                "name": "New User",
                "email": "oauth2@example.com",
                "avatar": None,
            },
        )

        assert filtered_data == {
            "id": "".join(reversed(str(user.id))),
            "name": "resU weN",
            "email": "filtered_oauth2@example.com",
            "avatar": None,
        }


def test_original_user_data_is_not_mutated_by_default_filters(user, request):
    user_data = {
        "id": "1234",
        "name": "New User",
        "email": "oauth2@example.com",
        "avatar": None,
    }

    filtered_data = filter_user_data(request, user, user_data)
    assert filtered_data == {
        "id": "1234",
        "name": "New_User",
        "email": "oauth2@example.com",
        "avatar": None,
    }

    assert user_data == {
        "id": "1234",
        "name": "New User",
        "email": "oauth2@example.com",
        "avatar": None,
    }


def test_original_user_data_is_not_mutated_by_custom_filters(user, request):
    def user_is_set_filter(request, user_obj, user_data):
        assert user_obj == user
        return {
            "id": str(user_obj.id),
            "name": user_data["name"],
            "email": user_data["email"],
            "avatar": user_data["avatar"],
        }

    user_data = {
        "id": "1234",
        "name": "New User",
        "email": "oauth2@example.com",
        "avatar": None,
    }

    with patch(
        "misago.oauth2.validation.oauth2_user_data_filters",
        [
            user_request_filter,
            user_is_set_filter,
            user_id_filter,
            user_name_filter,
            user_email_filter,
        ],
    ):
        filtered_data = filter_user_data(request, user, user_data)

        assert filtered_data == {
            "id": "".join(reversed(str(user.id))),
            "name": "resU weN",
            "email": "filtered_oauth2@example.com",
            "avatar": None,
        }

    assert user_data == {
        "id": "1234",
        "name": "New User",
        "email": "oauth2@example.com",
        "avatar": None,
    }



from django.db.migrations import RunPython


class StartCacheVersioning(RunPython):
    def __init__(self, cache):
        code = start_cache_versioning(cache)
        reverse_code = stop_cache_versioning(cache)
        super().__init__(code, reverse_code)


class StopCacheVersioning(RunPython):
    def __init__(self, cache):
        code = stop_cache_versioning(cache)
        reverse_code = start_cache_versioning(cache)
        super().__init__(code, reverse_code)


def start_cache_versioning(cache):
    def migration_operation(apps, _):
        CacheVersion = apps.get_model("misago_cache", "CacheVersion")
        CacheVersion.objects.create(cache=cache)

    return migration_operation


def stop_cache_versioning(cache):
    def migration_operation(apps, _):
        CacheVersion = apps.get_model("misago_cache", "CacheVersion")
        CacheVersion.objects.filter(cache=cache).delete()

    return migration_operation



from django.templatetags.static import static
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect

from ...conf import settings
from ..models import Attachment, AttachmentType


def attachment_server(request, pk, secret, thumbnail=False):
    try:
        url = serve_file(request, pk, secret, thumbnail)
        return redirect(url)
    except PermissionDenied:
        error_image = request.settings.attachment_403_image
        if not error_image:
            error_image = static(settings.MISAGO_ATTACHMENT_403_IMAGE)
        return redirect(error_image)
    except Http404:
        error_image = request.settings.attachment_404_image
        if not error_image:
            error_image = static(settings.MISAGO_ATTACHMENT_404_IMAGE)
        return redirect(error_image)


def serve_file(request, pk, secret, thumbnail):
    queryset = Attachment.objects.select_related("filetype")
    attachment = get_object_or_404(queryset, pk=pk, secret=secret)

    if not attachment.post_id and request.GET.get("shva") != "1":
        # if attachment is orphaned, don't run acl test unless explicitly told so
        # this saves user suprise of deleted attachment still showing in posts/quotes
        raise Http404()

    if not request.user.is_staff:
        allow_file_download(request, attachment)

    if attachment.is_image:
        if thumbnail and attachment.thumbnail:
            return attachment.thumbnail.url
        return attachment.image.url

    if thumbnail:
        raise Http404()
    return attachment.file.url


def allow_file_download(request, attachment):
    is_authenticated = request.user.is_authenticated

    if not is_authenticated or request.user.id != attachment.uploader_id:
        if not attachment.post_id:
            raise Http404()
        if not request.user_acl["can_download_other_users_attachments"]:
            raise PermissionDenied()

    allowed_roles = set(r.pk for r in attachment.filetype.limit_downloads_to.all())
    if allowed_roles:
        user_roles = set(r.pk for r in request.user.get_roles())
        if not user_roles & allowed_roles:
            raise PermissionDenied()

    if attachment.filetype.status == AttachmentType.DISABLED:
        raise PermissionDenied()



from django.test import TestCase

from ..page import Page


class SiteTests(TestCase):
    def setUp(self):
        self.page = Page("test")

    def test_pages(self):
        """add_section adds section to page"""
        self.page.add_section(
            link="misago:user-posts", name="Posts", after="misago:user-threads"
        )

        self.page.add_section(link="misago:user-threads", name="Threads")

        self.page.add_section(
            link="misago:user-follows", name="Follows", before="misago:user-posts"
        )

        self.page.assert_is_finalized()

        sorted_sections = self.page._sorted_list
        self.assertEqual(sorted_sections[0]["name"], "Threads")
        self.assertEqual(sorted_sections[1]["name"], "Follows")
        self.assertEqual(sorted_sections[2]["name"], "Posts")

        self.assertEqual(self.page.get_default_link(), "misago:user-threads")



from django import forms
from django.utils.translation import gettext_lazy as _

from ....admin.forms import YesNoSwitch
from ....core.validators import validate_image_square
from ....users.validators import validate_username_content
from ... import settings
from .base import ChangeSettingsForm


class ChangeUsersSettingsForm(ChangeSettingsForm):
    settings = [
        "account_activation",
        "allow_custom_avatars",
        "avatar_upload_limit",
        "default_avatar",
        "default_gravatar_fallback",
        "blank_avatar",
        "signature_length_max",
        "subscribe_reply",
        "subscribe_start",
        "username_length_max",
        "username_length_min",
        "anonymous_username",
        "users_per_page",
        "users_per_page_orphans",
        "top_posters_ranking_length",
        "top_posters_ranking_size",
        "allow_data_downloads",
        "data_downloads_expiration",
        "allow_delete_own_account",
        "new_inactive_accounts_delete",
        "ip_storage_time",
    ]

    account_activation = forms.ChoiceField(
        label=_("Require new accounts activation"),
        choices=[
            ("none", _("No activation required")),
            ("user", _("Activation token sent to user e-mail")),
            ("admin", _("Activation by administrator")),
            ("closed", _("Disable new registrations")),
        ],
        widget=forms.RadioSelect(),
    )
    new_inactive_accounts_delete = forms.IntegerField(
        label=_(
            "Delete new inactive accounts if they weren't activated "
            "within this number of days"
        ),
        help_text=_("Enter 0 to never delete inactive new accounts."),
        min_value=0,
    )

    username_length_min = forms.IntegerField(
        label=_("Minimum allowed username length"), min_value=2, max_value=20
    )
    username_length_max = forms.IntegerField(
        label=_("Maximum allowed username length"), min_value=2, max_value=20
    )

    allow_custom_avatars = YesNoSwitch(
        label=_("Allow custom avatar uploads"),
        help_text=_(
            "Turning this option off will forbid forum users from uploading custom "
            "avatars. Good for forums adressed at young users."
        ),
    )
    avatar_upload_limit = forms.IntegerField(
        label=_("Maximum size of uploaded avatar"),
        help_text=_("Enter maximum allowed file size (in KB) for avatar uploads."),
        min_value=0,
    )
    default_avatar = forms.ChoiceField(
        label=_("Default avatar"),
        choices=[
            ("dynamic", _("Individual")),
            ("gravatar", _("Gravatar")),
            ("gallery", _("Random avatar from gallery")),
        ],
        widget=forms.RadioSelect(),
    )
    default_gravatar_fallback = forms.ChoiceField(
        label=_("Fallback for default gravatar"),
        help_text=_(
            "Select which avatar to use when user has no gravatar associated with "
            "their e-mail address."
        ),
        choices=[
            ("dynamic", _("Individual")),
            ("gallery", _("Random avatar from gallery")),
        ],
        widget=forms.RadioSelect(),
    )
    blank_avatar = forms.ImageField(
        label=_("Blank avatar"),
        help_text=_(
            "Blank avatar is displayed in the interface when user's avatar is not "
            "available: when user was deleted or is guest. Uploaded image should be "
            "a square."
        ),
        required=False,
    )
    blank_avatar_delete = forms.BooleanField(
        label=_("Delete custom blank avatar"), required=False
    )

    signature_length_max = forms.IntegerField(
        label=_("Maximum allowed signature length"), min_value=10, max_value=5000
    )

    subscribe_start = forms.ChoiceField(
        label=_("Started threads"),
        choices=[
            ("no", _("Don't watch")),
            ("watch", _("Put on watched threads list")),
            (
                "watch_email",
                _("Put on watched threads list and e-mail user when somebody replies"),
            ),
        ],
        widget=forms.RadioSelect(),
    )
    subscribe_reply = forms.ChoiceField(
        label=_("Replied threads"),
        choices=[
            ("no", _("Don't watch")),
            ("watch", _("Put on watched threads list")),
            (
                "watch_email",
                _("Put on watched threads list and e-mail user when somebody replies"),
            ),
        ],
        widget=forms.RadioSelect(),
    )

    users_per_page = forms.IntegerField(
        label=_("Number of users displayed on a single page"), min_value=4
    )
    users_per_page_orphans = forms.IntegerField(
        label=_("Maximum orphans"),
        help_text=_(
            "If number of users to be displayed on the last page is less or equal to "
            "number specified in this setting, those users will instead be displayed "
            "on previous page, reducing the total number of pages on the list."
        ),
        min_value=0,
    )

    top_posters_ranking_length = forms.IntegerField(
        label=_("Maximum age in days of posts that should count to the ranking"),
        min_value=1,
    )
    top_posters_ranking_size = forms.IntegerField(
        label=_("Maximum number of ranked users"), min_value=2
    )

    allow_data_downloads = YesNoSwitch(label=_("Allow users to download their data"))
    data_downloads_expiration = forms.IntegerField(
        label=_("Maximum age in hours of data downloads before they expire"),
        help_text=_(
            "Data downloads older than specified will have their files deleted and "
            "will be marked as expired."
        ),
        min_value=1,
    )

    allow_delete_own_account = YesNoSwitch(
        label=_("Allow users to delete their own accounts")
    )

    ip_storage_time = forms.IntegerField(
        label=_("IP storage time"),
        help_text=_(
            "Number of days for which users IP addresses are stored in forum database. "
            "Enter zero to store registered IP addresses forever. Deleting user "
            "account always deletes the IP addresses associated with it."
        ),
        min_value=0,
    )

    anonymous_username = forms.CharField(
        label=_("Anonymous username"),
        help_text=_(
            "This username is displayed instead of delete user's actual name "
            "next to their content."
        ),
        min_length=1,
        max_length=15,
        validators=[validate_username_content],
    )

    def clean_blank_avatar(self):
        upload = self.cleaned_data.get("blank_avatar")
        if not upload or upload == self.initial.get("blank_avatar"):
            return None

        validate_image_square(upload.image)
        min_size = max(settings.MISAGO_AVATARS_SIZES)
        if upload.image.width < min_size:
            raise forms.ValidationError(
                _("Uploaded image's edge should be at least %(size)s pixels long.")
                % {"size": min_size}
            )

        return upload

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("users_per_page_orphans") > cleaned_data.get(
            "users_per_page"
        ):
            self.add_error(
                "users_per_page_orphans",
                _("This value must be lower than number of users per page."),
            )
        return cleaned_data



from django.core.files.base import ContentFile

from ..activetheme import get_active_theme
from ..models import Theme


def test_active_theme_data_can_be_obtained(db):
    assert get_active_theme()


def test_if_active_theme_is_default_theme_include_defaults_flag_is_set(db):
    assert get_active_theme()["include_defaults"]


def test_if_active_theme_is_default_themes_child_include_defaults_flag_is_set(
    default_theme, active_theme
):
    active_theme.parent = default_theme
    active_theme.save()
    assert get_active_theme()["include_defaults"]


def test_if_active_theme_is_not_default_themes_child_include_defaults_flag_is_not_set(
    active_theme,
):
    assert not get_active_theme()["include_defaults"]


def test_active_theme_styles_are_included(active_theme):
    active_theme.css.create(name="test", url="https://example.com")
    assert get_active_theme()["styles"]


def test_active_theme_parents_styles_are_included(active_theme):
    parent_theme = Theme.objects.create(name="Parent theme")
    parent_theme.css.create(name="test", url="https://example.com")

    active_theme.move_to(parent_theme)
    active_theme.save()

    assert get_active_theme()["styles"]


def test_active_theme_child_themes_styles_are_not_included(active_theme):
    child_theme = Theme.objects.create(parent=active_theme, name="Child theme")
    child_theme.css.create(name="test", url="https://example.com")
    assert not get_active_theme()["styles"]


def test_active_theme_styles_are_ordered(active_theme):
    last_css = active_theme.css.create(
        name="test", url="https://last-example.com", order=1
    )
    first_css = active_theme.css.create(
        name="test", url="https://first-example.com", order=0
    )

    assert get_active_theme()["styles"] == [first_css.url, last_css.url]


def test_active_theme_styles_list_includes_url_to_remote_css(active_theme):
    css = active_theme.css.create(name="test", url="https://last-example.com")
    assert get_active_theme()["styles"] == [css.url]


def test_active_theme_styles_list_contains_url_to_local_css(active_theme):
    css = active_theme.css.create(
        name="test",
        source_file=ContentFile("body {}", name="test.css"),
        source_hash="abcdefgh",
    )
    assert get_active_theme()["styles"] == [css.source_file.url]


def test_active_theme_styles_list_contains_url_to_local_built_css(active_theme):
    css = active_theme.css.create(
        name="test",
        source_needs_building=True,
        build_file=ContentFile("body {}", name="test.css"),
        build_hash="abcdefgh",
    )
    assert get_active_theme()["styles"] == [css.build_file.url]


def test_active_theme_styles_list_exclude_url_to_css_that_has_not_been_built(
    active_theme,
):
    active_theme.css.create(
        name="test",
        source_file=ContentFile("body {}", name="test.css"),
        source_hash="abcdefgh",
        source_needs_building=True,
    )
    assert not get_active_theme()["styles"]



# Generated by Django 2.2.1 on 2019-06-16 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SocialAuthProvider",
            fields=[
                (
                    "provider",
                    models.CharField(max_length=30, primary_key=True, serialize=False),
                ),
                (
                    "button_text",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("button_color", models.CharField(blank=True, max_length=7, null=True)),
                (
                    "settings",
                    models.JSONField(default=dict),
                ),
                ("is_active", models.BooleanField(default=False)),
                ("order", models.IntegerField(default=0)),
            ],
            options={"ordering": ["order"]},
        )
    ]



from unittest.mock import Mock

from django.test import TestCase
from django.utils import timezone

from ...acl import useracl
from ...acl.objectacl import add_acl_to_obj
from ...categories.models import Category
from ...conftest import get_cache_versions
from ...users.test import create_test_user
from ..events import record_event
from ..models import Thread

cache_versions = get_cache_versions()


class EventsApiTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")

        datetime = timezone.now()

        self.category = Category.objects.all_categories()[:1][0]
        self.thread = Thread(
            category=self.category,
            started_on=datetime,
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=datetime,
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        self.thread.set_title("Test thread")
        self.thread.save()

        user_acl = useracl.get_user_acl(self.user, cache_versions)
        add_acl_to_obj(user_acl, self.category)
        add_acl_to_obj(user_acl, self.thread)

    def test_record_event_with_context(self):
        """record_event registers event with context in thread"""
        request = Mock(user=self.user, user_ip="123.14.15.222")
        context = {"user": "Lorem ipsum"}
        event = record_event(request, self.thread, "announcement", context)

        event_post = self.thread.post_set.order_by("-id")[:1][0]
        self.assertEqual(self.thread.last_post, event_post)
        self.assertTrue(self.thread.has_events)
        self.assertTrue(self.thread.last_post_is_event)

        self.assertEqual(event.pk, event_post.pk)
        self.assertTrue(event_post.is_event)
        self.assertEqual(event_post.event_type, "announcement")
        self.assertEqual(event_post.event_context, context)
        self.assertEqual(event_post.poster_id, request.user.pk)

    def test_record_event_is_read(self):
        """record_event makes recorded event read to its author"""
        request = Mock(user=self.user, user_ip="123.14.15.222")
        event = record_event(request, self.thread, "announcement")

        self.user.postread_set.get(
            category=self.category, thread=self.thread, post=event
        )



import responses

from ..tasks import update_remote_css_size


def mock_http_head_response(css_link, content_length=None):
    headers = {}
    if content_length is not None:
        headers["content-length"] = str(content_length)

    responses.add(responses.HEAD, css_link.url, headers=headers)


@responses.activate
def test_task_uses_content_length_header_to_set_css_size(css_link):
    content_length = 343
    mock_http_head_response(css_link, content_length)

    update_remote_css_size(css_link.pk)

    css_link.refresh_from_db()
    assert css_link.size == content_length


@responses.activate
def test_task_doesnt_change_css_size_if_content_length_header_is_incorrect(css_link):
    mock_http_head_response(css_link, "incorrect")
    update_remote_css_size(css_link.pk)

    css_link.refresh_from_db()
    assert css_link.size == 0


@responses.activate
def test_task_doesnt_change_css_size_if_content_length_header_is_empty(css_link):
    mock_http_head_response(css_link, "")
    update_remote_css_size(css_link.pk)

    css_link.refresh_from_db()
    assert css_link.size == 0


@responses.activate
def test_task_doesnt_change_css_size_if_content_length_header_is_not_set(css_link):
    mock_http_head_response(css_link)
    update_remote_css_size(css_link.pk)

    css_link.refresh_from_db()
    assert css_link.size == 0


@responses.activate
def test_task_doesnt_change_css_size_if_http_request_failed(css_link):
    responses.add(responses.HEAD, css_link.url, status=404)
    update_remote_css_size(css_link.pk)


def test_task_handles_css_without_url(css):
    update_remote_css_size(css.pk)


def test_task_handles_nonexisting_css(db):
    update_remote_css_size(1)



from unittest.mock import call

import pytest
from django.contrib.auth import get_user_model
from django.core import mail

from ... import BANS_CACHE
from ....cache.test import assert_invalidates_cache
from ....test import assert_contains, assert_has_error_message, assert_not_contains
from ...datadownloads import request_user_data_download
from ...models import Ban, DataDownload, DeletedUser
from ...test import create_test_user

User = get_user_model()


def create_multiple_users(**kwargs):
    return [
        create_test_user("User%s" % i, "user%s@gmail.com" % i, **kwargs)
        for i in range(5)
    ]


def get_multiple_users_ids(**kwargs):
    users = create_multiple_users(**kwargs)
    return [u.id for u in users]


@pytest.fixture
def users_ids(db):
    return get_multiple_users_ids()


def test_multiple_users_can_be_activated_with_mass_action(
    admin_client, users_admin_link
):
    users = create_multiple_users(requires_activation=1)
    response = admin_client.post(
        users_admin_link,
        data={"action": "activate", "selected_items": [u.id for u in users]},
    )

    for user in users:
        user.refresh_from_db()
        assert not user.requires_activation


def test_activating_multiple_users_sends_email_notifications_to_them(
    admin_client, users_admin_link
):
    users_ids = get_multiple_users_ids(requires_activation=1)
    response = admin_client.post(
        users_admin_link, data={"action": "activate", "selected_items": users_ids}
    )

    assert len(mail.outbox) == len(users_ids)
    assert "has been activated" in mail.outbox[0].subject


def test_ban_multiple_users_form_is_rendered(admin_client, users_admin_link):
    users_ids = get_multiple_users_ids()
    response = admin_client.post(
        users_admin_link, data={"action": "ban", "selected_items": users_ids}
    )
    assert response.status_code == 200


def test_multiple_users_can_be_banned_with_mass_action(admin_client, users_admin_link):
    users = create_multiple_users()
    admin_client.post(
        users_admin_link,
        data={
            "action": "ban",
            "selected_items": [u.id for u in users],
            "ban_type": ["usernames", "emails", "domains"],
            "finalize": "",
        },
    )

    for user in users:
        Ban.objects.get(banned_value=user.username.lower())
        Ban.objects.get(banned_value=user.email)
        Ban.objects.get(banned_value="*%s" % user.email[-10:])


def test_option_to_ban_multiple_users_ips_is_disabled_if_user_ips_are_not_available(
    admin_client, users_admin_link, users_ids
):
    response = admin_client.post(
        users_admin_link, data={"action": "ban", "selected_items": users_ids}
    )
    assert_not_contains(response, 'value="ip"')
    assert_not_contains(response, 'value="ip_first"')
    assert_not_contains(response, 'value="ip_two"')


def test_option_to_ban_multiple_users_ips_is_enabled_if_user_ips_are_available(
    admin_client, users_admin_link
):
    users_ids = get_multiple_users_ids(joined_from_ip="1.2.3.4")
    response = admin_client.post(
        users_admin_link, data={"action": "ban", "selected_items": users_ids}
    )
    assert_contains(response, 'value="ip"')
    assert_contains(response, 'value="ip_first"')
    assert_contains(response, 'value="ip_two"')


def test_multiple_users_ips_can_be_banned_with_mass_action(
    admin_client, users_admin_link
):
    users_ids = get_multiple_users_ids(joined_from_ip="1.2.3.4")
    response = admin_client.post(
        users_admin_link,
        data={
            "action": "ban",
            "selected_items": users_ids,
            "ban_type": ["ip", "ip_first", "ip_two"],
            "finalize": "",
        },
    )

    Ban.objects.get(banned_value="1.2.3.4")
    Ban.objects.get(banned_value="1.*")
    Ban.objects.get(banned_value="1.2.*")


def test_banning_multiple_users_with_mass_action_invalidates_bans_cache(
    admin_client, users_admin_link, users_ids
):
    with assert_invalidates_cache(BANS_CACHE):
        admin_client.post(
            users_admin_link,
            data={
                "action": "ban",
                "selected_items": users_ids,
                "ban_type": ["usernames", "emails", "domains"],
                "finalize": "",
            },
        )


def test_data_downloads_can_be_requested_for_multiple_users_with_mass_action(
    admin_client, users_admin_link
):
    users = create_multiple_users()
    response = admin_client.post(
        users_admin_link,
        data={
            "action": "request_data_download",
            "selected_items": [u.id for u in users],
        },
    )

    for user in users:
        DataDownload.objects.get(user=user)


def test_mass_action_is_not_requesting_data_downloads_for_users_with_existing_requests(
    admin_client, users_admin_link
):
    users = create_multiple_users()
    downloads_ids = [request_user_data_download(u).id for u in users]

    response = admin_client.post(
        users_admin_link,
        data={
            "action": "request_data_download",
            "selected_items": [u.id for u in users],
        },
    )

    assert not DataDownload.objects.exclude(id__in=downloads_ids).exists()


def test_multiple_users_can_be_deleted_with_mass_action(admin_client, users_admin_link):
    users = create_multiple_users()
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_accounts", "selected_items": [u.id for u in users]},
    )

    for user in users:
        with pytest.raises(User.DoesNotExist):
            user.refresh_from_db()


def test_mass_action_records_users_deletion_by_staff(admin_client, users_admin_link):
    users = create_multiple_users()
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_accounts", "selected_items": [u.id for u in users]},
    )

    deleted_count = DeletedUser.objects.filter(
        deleted_by=DeletedUser.DELETED_BY_STAFF
    ).count()
    assert deleted_count == len(users)


def test_delete_users_mass_action_fails_if_user_tries_to_delete_themselves(
    admin_client, users_admin_link, superuser
):
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_accounts", "selected_items": [superuser.id]},
    )
    assert_has_error_message(response)
    superuser.refresh_from_db()


def test_delete_users_mass_action_fails_if_user_tries_to_delete_staff_members(
    admin_client, users_admin_link
):
    users = create_multiple_users(is_staff=True)
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_accounts", "selected_items": [u.id for u in users]},
    )
    assert_has_error_message(response)

    for user in users:
        user.refresh_from_db()


def test_delete_users_mass_action_fails_if_user_tries_to_delete_superusers(
    admin_client, users_admin_link
):
    users = create_multiple_users(is_superuser=True)
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_accounts", "selected_items": [u.id for u in users]},
    )
    assert_has_error_message(response)

    for user in users:
        user.refresh_from_db()


@pytest.fixture
def mock_delete_user_with_content(mocker):
    delay = mocker.Mock()
    mocker.patch(
        "misago.users.admin.views.users.delete_user_with_content",
        mocker.Mock(delay=delay),
    )
    return delay


def test_multiple_users_can_be_deleted_together_with_content_by_mass_action(
    admin_client, users_admin_link, users_ids, mock_delete_user_with_content
):
    response = admin_client.post(
        users_admin_link, data={"action": "delete_all", "selected_items": users_ids}
    )

    calls = [call(u) for u in users_ids]
    mock_delete_user_with_content.assert_has_calls(calls, any_order=True)


def test_deleting_multiple_users_with_content_disables_their_accounts(
    admin_client, users_admin_link, mock_delete_user_with_content
):
    users = create_multiple_users()
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_all", "selected_items": [u.id for u in users]},
    )

    for user in users:
        user.refresh_from_db()
        assert not user.is_active


def test_delete_users_with_content_mass_action_fails_if_user_tries_to_delete_themselves(
    admin_client, users_admin_link, superuser, mock_delete_user_with_content
):
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_all", "selected_items": [superuser.id]},
    )

    assert_has_error_message(response)
    mock_delete_user_with_content.assert_not_called()

    superuser.refresh_from_db()
    assert superuser.is_active


def test_delete_users_with_content_mass_action_fails_if_user_tries_to_delete_staff(
    admin_client, users_admin_link, mock_delete_user_with_content
):
    users = create_multiple_users(is_staff=True)
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_all", "selected_items": [u.id for u in users]},
    )

    assert_has_error_message(response)
    mock_delete_user_with_content.assert_not_called()

    for user in users:
        user.refresh_from_db()
        assert user.is_active


def test_delete_users_with_content_mass_action_fails_if_user_tries_to_delete_superusers(
    admin_client, users_admin_link, mock_delete_user_with_content
):
    users = create_multiple_users(is_superuser=True)
    response = admin_client.post(
        users_admin_link,
        data={"action": "delete_all", "selected_items": [u.id for u in users]},
    )

    assert_has_error_message(response)
    mock_delete_user_with_content.assert_not_called()

    for user in users:
        user.refresh_from_db()
        assert user.is_active



# Generated by Django 1.10.5 on 2017-02-05 14:34
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0005_dj_19_update")]

    operations = []



from django.urls import reverse
from rest_framework import serializers

from ..models import PostLike

__all__ = ["PostLikeSerializer"]


class PostLikeSerializer(serializers.ModelSerializer):
    avatars = serializers.SerializerMethodField()
    liker_id = serializers.SerializerMethodField()
    username = serializers.SerializerMethodField()

    url = serializers.SerializerMethodField()

    class Meta:
        model = PostLike
        fields = ["id", "avatars", "liked_on", "liker_id", "username", "url"]

    def get_liker_id(self, obj):
        return obj["liker_id"]

    def get_username(self, obj):
        return obj["liker_name"]

    def get_avatars(self, obj):
        return obj.get("liker__avatars")

    def get_url(self, obj):
        if obj["liker_id"]:
            return reverse(
                "misago:user", kwargs={"slug": obj["liker_slug"], "pk": obj["liker_id"]}
            )



from ..threads.models import Post, Thread
from ..threads.permissions import exclude_invisible_posts, exclude_invisible_threads
from .cutoffdate import get_cutoff_date


def make_read_aware(request, categories):
    if not categories:
        return

    if not hasattr(categories, "__iter__"):
        categories = [categories]

    make_read(categories)

    if request.user.is_anonymous:
        return

    threads = Thread.objects.filter(category__in=categories)
    threads = exclude_invisible_threads(request.user_acl, categories, threads)

    queryset = (
        Post.objects.filter(
            category__in=categories,
            thread__in=threads,
            posted_on__gt=get_cutoff_date(request.settings, request.user),
        )
        .values_list("category", flat=True)
        .distinct()
    )

    queryset = queryset.exclude(id__in=request.user.postread_set.values("post"))
    queryset = exclude_invisible_posts(request.user_acl, categories, queryset)

    unread_categories = list(queryset)

    for category in categories:
        if category.pk in unread_categories:
            category.is_read = False
            category.is_new = True


def make_read(threads):
    for thread in threads:
        thread.is_read = True
        thread.is_new = False



import os

from django.conf import settings
from django.test import TestCase
from django.urls import reverse
from django.utils import translation

MISAGO_DIR = os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
)
LOCALES_DIR = os.path.join(MISAGO_DIR, "locale")


class JsI18nUrlTests(TestCase):
    def test_url_cache_buster(self):
        """js i18n catalog link has cachebuster with lang code"""
        url = "%s?%s" % (reverse("django-i18n"), settings.LANGUAGE_CODE)

        response = self.client.get(reverse("misago:index"))
        self.assertContains(response, url)

    def test_js_catalogs_are_correct(self):
        """no JS catalogs have showstoppers"""
        failed_languages = []
        for language in os.listdir(LOCALES_DIR):
            if "." in language:
                continue
            try:
                with translation.override(language):
                    response = self.client.get(reverse("django-i18n"))
                    if response.status_code != 200:
                        failed_languages.append(language)
            except Exception:  # pylint: disable=broad-except
                failed_languages.append(language)

        if failed_languages:
            self.fail(
                "JS catalog failed for languages: %s" % (", ".join(failed_languages))
            )



from django.contrib.auth import get_user_model
from django.urls import reverse

from .. import test
from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase
from ..test import patch_category_acl

User = get_user_model()


class SubscriptionMiddlewareTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.category = Category.objects.get(slug="first-category")


class SubscribeStartedThreadTests(SubscriptionMiddlewareTestCase):
    def setUp(self):
        super().setUp()
        self.api_link = reverse("misago:api:thread-list")

    @patch_category_acl({"can_start_threads": True})
    def test_dont_subscribe(self):
        """middleware makes no subscription to thread"""
        self.user.subscribe_to_started_threads = User.SUBSCRIPTION_NONE
        self.user.subscribe_to_replied_threads = User.SUBSCRIPTION_NOTIFY
        self.user.save()

        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.id,
                "title": "This is an test thread!",
                "post": "This is test response!",
            },
        )
        self.assertEqual(response.status_code, 200)

        # user has no subscriptions
        self.assertEqual(self.user.subscription_set.count(), 0)

    @patch_category_acl({"can_start_threads": True})
    def test_subscribe(self):
        """middleware subscribes thread"""
        self.user.subscribe_to_started_threads = User.SUBSCRIPTION_NOTIFY
        self.user.save()

        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.id,
                "title": "This is an test thread!",
                "post": "This is test response!",
            },
        )
        self.assertEqual(response.status_code, 200)

        # user has subscribed to thread
        thread = self.user.thread_set.order_by("id").last()
        subscription = self.user.subscription_set.get(thread=thread)

        self.assertEqual(subscription.category_id, self.category.id)
        self.assertFalse(subscription.send_email)

    @patch_category_acl({"can_start_threads": True})
    def test_email_subscribe(self):
        """middleware subscribes thread with an email"""
        self.user.subscribe_to_started_threads = User.SUBSCRIPTION_ALL
        self.user.save()

        response = self.client.post(
            self.api_link,
            data={
                "category": self.category.id,
                "title": "This is an test thread!",
                "post": "This is test response!",
            },
        )
        self.assertEqual(response.status_code, 200)

        # user has subscribed to thread
        thread = self.user.thread_set.order_by("id").last()
        subscription = self.user.subscription_set.get(thread=thread)

        self.assertEqual(subscription.category_id, self.category.id)
        self.assertTrue(subscription.send_email)


class SubscribeRepliedThreadTests(SubscriptionMiddlewareTestCase):
    def setUp(self):
        super().setUp()
        self.thread = test.post_thread(self.category)
        self.api_link = reverse(
            "misago:api:thread-post-list", kwargs={"thread_pk": self.thread.pk}
        )

    @patch_category_acl({"can_reply_threads": True})
    def test_dont_subscribe(self):
        """middleware makes no subscription to thread"""
        self.user.subscribe_to_started_threads = User.SUBSCRIPTION_NOTIFY
        self.user.subscribe_to_replied_threads = User.SUBSCRIPTION_NONE
        self.user.save()

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        # user has no subscriptions
        self.assertEqual(self.user.subscription_set.count(), 0)

    @patch_category_acl({"can_reply_threads": True})
    def test_subscribe(self):
        """middleware subscribes thread"""
        self.user.subscribe_to_replied_threads = User.SUBSCRIPTION_NOTIFY
        self.user.save()

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        # user has subscribed to thread
        subscription = self.user.subscription_set.get(thread=self.thread)

        self.assertEqual(subscription.category_id, self.category.id)
        self.assertFalse(subscription.send_email)

    @patch_category_acl({"can_reply_threads": True})
    def test_email_subscribe(self):
        """middleware subscribes thread with an email"""
        self.user.subscribe_to_replied_threads = User.SUBSCRIPTION_ALL
        self.user.save()

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        # user has subscribed to thread
        subscription = self.user.subscription_set.get(thread=self.thread)

        self.assertEqual(subscription.category_id, self.category.id)
        self.assertTrue(subscription.send_email)

    @patch_category_acl({"can_reply_threads": True})
    def test_subscribe_with_events(self):
        """middleware omits events when testing for replied thread"""
        self.user.subscribe_to_replied_threads = User.SUBSCRIPTION_ALL
        self.user.save()

        # set event in thread
        test.reply_thread(self.thread, self.user, is_event=True)

        # reply thread
        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        # user has subscribed to thread
        subscription = self.user.subscription_set.get(thread=self.thread)

        self.assertEqual(subscription.category_id, self.category.id)
        self.assertTrue(subscription.send_email)

    @patch_category_acl({"can_reply_threads": True})
    @patch_user_acl({"can_omit_flood_protection": True})
    def test_dont_subscribe_replied(self):
        """middleware omits threads user already replied"""
        self.user.subscribe_to_replied_threads = User.SUBSCRIPTION_ALL
        self.user.save()

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        # clear subscription
        self.user.subscription_set.all().delete()

        # reply again
        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        # user has no subscriptions
        self.assertEqual(self.user.subscription_set.count(), 0)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_multi_line_quote 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1"></div>
<blockquote class="quote-body" data-block="quote">
<p>Sit amet elit.</p>
<p>Another line.</p>
</blockquote>
</aside>"""

snapshots[
    "test_quote_can_contain_bbcode_or_markdown 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1"></div>
<blockquote class="quote-body" data-block="quote">
<p>Sit <strong>amet</strong> <u>elit</u>.</p>
</blockquote>
</aside>"""

snapshots[
    "test_quotes_can_be_nested 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1"></div>
<blockquote class="quote-body" data-block="quote">
<p>Sit amet elit.</p>
<aside class="quote-block">
<div class="quote-heading" data-noquote="1"></div>
<blockquote class="quote-body" data-block="quote">
<p>Nested quote</p>
</blockquote>
</aside>
</blockquote>
</aside>"""

snapshots[
    "test_quotes_can_contain_hr_markdown 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1"></div>
<blockquote class="quote-body" data-block="quote">
<p>Sit amet elit.</p>
<hr />
<p>Another line.</p>
</blockquote>
</aside>"""

snapshots[
    "test_single_line_authored_quote 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1">@Bob</div>
<blockquote class="quote-body" data-author="@Bob" data-block="quote">
<p>Sit amet elit.</p>
</blockquote>
</aside>"""

snapshots[
    "test_single_line_authored_quote_without_quotations 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1">@Bob</div>
<blockquote class="quote-body" data-author="@Bob" data-block="quote">
<p>Sit amet elit.</p>
</blockquote>
</aside>"""

snapshots[
    "test_single_line_quote 1"
] = """<aside class="quote-block">
<div class="quote-heading" data-noquote="1"></div>
<blockquote class="quote-body" data-block="quote">
<p>Sit amet elit.</p>
</blockquote>
</aside>"""



from dataclasses import dataclass

from django.contrib.auth import get_user_model
from django.core.validators import validate_email
from django.forms import ValidationError
from django.utils.crypto import get_random_string
from unidecode import unidecode

from ..hooks import oauth2_validators, oauth2_user_data_filters
from ..users.validators import (
    validate_username_content,
    validate_username_length,
)
from .exceptions import OAuth2UserDataValidationError

User = get_user_model()


class UsernameSettings:
    username_length_max: int = 200
    username_length_min: int = 1


def validate_user_data(request, user, user_data, response_json):
    filtered_data = filter_user_data(request, user, user_data)

    try:
        validate_username_content(filtered_data["name"])
        validate_username_length(UsernameSettings, filtered_data["name"])
        validate_email(filtered_data["email"])

        for plugin_oauth2_validator in oauth2_validators:
            plugin_oauth2_validator(request, user, user_data, response_json)
    except ValidationError as exc:
        raise OAuth2UserDataValidationError(error_list=[str(exc.message)])

    return filtered_data


def filter_user_data(request, user, user_data):
    filtered_data = {
        "id": user_data["id"],
        "name": str(user_data["name"] or "").strip(),
        "email": str(user_data["email"] or "").strip(),
        "avatar": filter_user_avatar(user_data["avatar"]),
    }

    if oauth2_user_data_filters:
        return filter_user_data_with_filters(
            request, user, filtered_data, oauth2_user_data_filters
        )
    else:
        filtered_data["name"] = filter_name(user, filtered_data["name"])

    return filtered_data


def filter_user_avatar(user_avatar):
    if user_avatar:
        return str(user_avatar).strip() or None
    return None


def filter_user_data_with_filters(request, user, user_data, filters):
    for filter_user_data in filters:
        user_data = filter_user_data(request, user, user_data.copy()) or user_data
    return user_data


def filter_name(user, name):
    if user and user.username == name:
        return name

    clean_name = "".join(
        [c for c in unidecode(name.replace(" ", "_")) if c.isalnum() or c == "_"]
    )

    if user and user.username == clean_name:
        return clean_name  # No change in name

    if not clean_name.replace("_", ""):
        clean_name = "User_%s" % get_random_string(4)

    clean_name_root = clean_name
    while True:
        try:
            db_user = User.objects.get_by_username(clean_name)
        except User.DoesNotExist:
            return clean_name

        if not user or user.pk != db_user.pk:
            clean_name = f"{clean_name_root}_{get_random_string(4)}"
        else:
            return clean_name



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-11 20:51
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0017_move_bans_to_cache_version")]

    operations = [
        migrations.AlterField(
            model_name="datadownload",
            name="requested_on",
            field=models.DateTimeField(
                db_index=True, default=django.utils.timezone.now
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="joined_on",
            field=models.DateTimeField(
                db_index=True,
                default=django.utils.timezone.now,
                verbose_name="joined on",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
import os

from setuptools import setup, find_packages

from misago import __version__ as version


SETUP_DIR = os.path.dirname(__file__)

with open(os.path.join(SETUP_DIR, "README.rst"), "rb") as f:
    README = f.read().decode("utf-8")

with open(os.path.join(SETUP_DIR, "requirements.txt"), "r") as f:
    REQUIREMENTS = f.read()


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="Misago",
    version=version,
    license="GPLv2",
    description=(
        "Misago is modern, fully featured forum application written in "
        "Python and ES6, powered by Django and React.js. It works out of "
        "the box and can be run alone or be connected to existing site "
        "with built in OAuth 2 client."
    ),
    long_description=README,
    url="http://www.misago-project.org/",
    author=u"Rafał Pitoń",
    author_email="kontakt@rpiton.com",
    install_requires=REQUIREMENTS,
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from .parser import parse


def common(request, poster, text, allow_mentions=True, force_shva=False):
    """
    Common flavour

    Used in places where full Misago flavour is desired

    Breaks text into paragraphs, supports code, spoiler and quote blocks,
    headers, lists, images, spoilers, text styles

    Returns dict object
    """
    return parse(
        text, request, poster, allow_mentions=allow_mentions, force_shva=force_shva
    )


def limited(request, text):
    """
    Limited flavour

    Breaks text in paragraphs, supports strong, em, i, u, b,
    automatically linkifies links.

    Returns parsed text
    """
    result = parse(
        text,
        request,
        request.user,
        allow_mentions=False,
        allow_links=True,
        allow_images=False,
        allow_blocks=False,
    )

    return result["parsed_text"]


def signature(request, owner, user_acl, text):
    result = parse(
        text,
        request,
        owner,
        allow_mentions=False,
        allow_blocks=user_acl["allow_signature_blocks"],
        allow_links=user_acl["allow_signature_links"],
        allow_images=user_acl["allow_signature_images"],
    )

    return result["parsed_text"]



from django.urls import reverse

ANONYMIZABLE_EVENTS = (
    "added_participant",
    "changed_owner",
    "owner_left",
    "removed_owner",
    "participant_left",
    "removed_participant",
)


def anonymize_event(user, event):
    if event.event_type not in ANONYMIZABLE_EVENTS:
        raise ValueError('event of type "%s" can\'t be ananymized' % event.event_type)

    event.event_context = {
        "user": {"id": None, "username": user.username, "url": reverse("misago:index")}
    }
    event.save(update_fields=["event_context"])


def anonymize_post_last_likes(user, post):
    cleaned_likes = []
    for like in post.last_likes:
        if like["id"] == user.id:
            cleaned_likes.append({"id": None, "username": user.username})
        else:
            cleaned_likes.append(like)

    if cleaned_likes != post.last_likes:
        post.last_likes = cleaned_likes
        post.save(update_fields=["last_likes"])



from rest_framework.response import Response

from ...serializers import PostLikeSerializer


def likes_list_endpoint(request, post):
    queryset = post.postlike_set.select_related("liker").values(
        "id", "liker_id", "liker_name", "liker_slug", "liked_on", "liker__avatars"
    )

    likes = []
    for like in queryset.iterator():
        likes.append(PostLikeSerializer(like).data)

    return Response(likes)



import json
from unittest.mock import Mock

from django.contrib.auth import get_user_model
from django.core import mail
from django.test import RequestFactory
from social_core.backends.github import GithubOAuth2
from social_django.utils import load_strategy

from ...acl.useracl import get_user_acl
from ...conf.dynamicsettings import DynamicSettings
from ...conf.test import override_dynamic_settings
from ...conftest import get_cache_versions
from ...core.exceptions import SocialAuthBanned, SocialAuthFailed
from ...legal.models import Agreement
from ...users.models import AnonymousUser, Ban, BanCache
from ...users.test import UserTestCase
from ..pipeline import (
    associate_by_email,
    create_user,
    create_user_with_form,
    get_username,
    require_activation,
    validate_ip_not_banned,
    validate_user_not_banned,
)

User = get_user_model()


def create_request(user_ip="0.0.0.0", data=None):
    factory = RequestFactory()
    if data is None:
        request = factory.get("/")
    else:
        request = factory.post(
            "/", data=json.dumps(data), content_type="application/json"
        )
    request.include_frontend_context = True
    request.cache_versions = get_cache_versions()
    request.frontend_context = {}
    request.socialauth = {}
    request.session = {}
    request.settings = DynamicSettings(request.cache_versions)
    request.user = AnonymousUser()
    request.user_acl = get_user_acl(request.user, request.cache_versions)
    request.user_ip = user_ip
    return request


def create_strategy():
    request = create_request()
    return load_strategy(request=request)


class MockStrategy:
    def __init__(self, user_ip="0.0.0.0"):
        self.cleaned_partial_token = None
        self.request = create_request(user_ip)

    def clean_partial_pipeline(self, token):
        self.cleaned_partial_token = token


class PipelineTestCase(UserTestCase):
    def get_initial_user(self):
        self.user = self.get_authenticated_user()

    def assertNewUserIsCorrect(
        self, new_user, form_data=None, activation=None, email_verified=False
    ):
        self.assertFalse(new_user.has_usable_password())
        self.assertIn("Welcome", mail.outbox[0].subject)

        if form_data:
            self.assertEqual(new_user.email, form_data["email"])
            self.assertEqual(new_user.username, form_data["username"])

        if activation == "none":
            self.assertEqual(new_user.requires_activation, User.ACTIVATION_NONE)

        if activation == "user":
            if email_verified:
                self.assertEqual(new_user.requires_activation, User.ACTIVATION_NONE)
            else:
                self.assertEqual(new_user.requires_activation, User.ACTIVATION_USER)

        if activation == "admin":
            self.assertEqual(new_user.requires_activation, User.ACTIVATION_ADMIN)

        self.assertEqual(new_user.audittrail_set.count(), 1)

    def assertJsonResponseEquals(self, response, value):
        response_content = response.content.decode("utf-8")
        response_json = json.loads(response_content)
        self.assertEqual(response_json, value)


class AssociateByEmailTests(PipelineTestCase):
    def test_skip_if_user_is_already_set(self):
        """pipeline step is skipped if user was found by previous step"""
        result = associate_by_email(Mock(), {}, GithubOAuth2, self.user)
        self.assertIsNone(result)

    def test_skip_if_no_email_passed(self):
        """pipeline step is skipped if no email was passed"""
        result = associate_by_email(Mock(), {}, GithubOAuth2)
        self.assertIsNone(result)

    def test_skip_if_user_with_email_not_found(self):
        """pipeline step is skipped if no email was passed"""
        result = associate_by_email(Mock(), {"email": "not@found.com"}, GithubOAuth2)
        self.assertIsNone(result)

    def test_raise_if_user_is_inactive(self):
        """pipeline raises if user was inactive"""
        strategy = Mock(setting=Mock(return_value=True))
        self.user.is_active = False
        self.user.save()

        try:
            associate_by_email(strategy, {"email": self.user.email}, GithubOAuth2)
            self.fail("associate_by_email should raise SocialAuthFailed")
        except SocialAuthFailed as e:
            self.assertEqual(
                e.message,
                (
                    "The e-mail address associated with your GitHub account "
                    "is not available for use on this site."
                ),
            )

    def test_raise_if_user_needs_admin_activation(self):
        """pipeline raises if user needs admin activation"""
        strategy = Mock(setting=Mock(return_value=True))
        self.user.requires_activation = User.ACTIVATION_ADMIN
        self.user.save()

        try:
            associate_by_email(strategy, {"email": self.user.email}, GithubOAuth2)
            self.fail("associate_by_email should raise SocialAuthFailed")
        except SocialAuthFailed as e:
            self.assertEqual(
                e.message,
                (
                    "Your account has to be activated by site administrator "
                    "before you will be able to sign in with GitHub."
                ),
            )

    def test_no_user_is_returned_if_pipeline_is_disabled(self):
        strategy = Mock(setting=Mock(return_value=False))
        result = associate_by_email(strategy, {"email": self.user.email}, GithubOAuth2)
        self.assertIsNone(result)
        strategy.setting.assert_called_once_with(
            "ASSOCIATE_BY_EMAIL", default=False, backend=GithubOAuth2
        )

    def test_return_user(self):
        """pipeline returns user if email was found"""
        strategy = Mock(setting=Mock(return_value=True))
        result = associate_by_email(strategy, {"email": self.user.email}, GithubOAuth2)
        self.assertEqual(result, {"user": self.user, "is_new": False})

    def test_return_user_email_inactive(self):
        """pipeline returns user even if they didn't activate their account manually"""
        strategy = Mock(setting=Mock(return_value=True))
        self.user.requires_activation = User.ACTIVATION_USER
        self.user.save()

        result = associate_by_email(strategy, {"email": self.user.email}, GithubOAuth2)
        self.assertEqual(result, {"user": self.user, "is_new": False})


class CreateUser(PipelineTestCase):
    def test_skip_if_user_is_set(self):
        """pipeline step is skipped if user was passed"""
        result = create_user(MockStrategy(), {}, GithubOAuth2(), user=self.user)
        self.assertIsNone(result)

    def test_skip_if_no_email_passed(self):
        """pipeline step is skipped if no email was passed"""
        result = create_user(
            MockStrategy(), {}, GithubOAuth2(), clean_username="TestBob"
        )
        self.assertIsNone(result)

    def test_skip_if_no_clean_username_passed(self):
        """pipeline step is skipped if cleaned username wasnt passed"""
        result = create_user(
            MockStrategy(), {"email": "hello@example.com"}, GithubOAuth2()
        )
        self.assertIsNone(result)

    def test_skip_if_email_is_taken(self):
        """pipeline step is skipped if email was taken"""
        result = create_user(
            MockStrategy(),
            {"email": self.user.email},
            GithubOAuth2(),
            clean_username="NewUser",
        )
        self.assertIsNone(result)

    @override_dynamic_settings(account_activation="none")
    def test_user_created_no_activation(self):
        """pipeline step creates active user for valid data and disabled activation"""
        result = create_user(
            MockStrategy(),
            {"email": "new@example.com"},
            GithubOAuth2(),
            clean_username="NewUser",
        )
        new_user = User.objects.get(email="new@example.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})
        self.assertEqual(new_user.username, "NewUser")
        self.assertNewUserIsCorrect(new_user, email_verified=True, activation="none")

    @override_dynamic_settings(account_activation="user")
    def test_user_created_activation_by_user(self):
        """pipeline step creates active user for valid data and user activation"""
        result = create_user(
            MockStrategy(),
            {"email": "new@example.com"},
            GithubOAuth2(),
            clean_username="NewUser",
        )
        new_user = User.objects.get(email="new@example.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})
        self.assertEqual(new_user.username, "NewUser")
        self.assertNewUserIsCorrect(new_user, email_verified=True, activation="user")

    @override_dynamic_settings(account_activation="admin")
    def test_user_created_activation_by_admin(self):
        """pipeline step creates in user for valid data and admin activation"""
        result = create_user(
            MockStrategy(),
            {"email": "new@example.com"},
            GithubOAuth2(),
            clean_username="NewUser",
        )
        new_user = User.objects.get(email="new@example.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})
        self.assertEqual(new_user.username, "NewUser")
        self.assertNewUserIsCorrect(new_user, email_verified=True, activation="admin")


class CreateUserWithFormTests(PipelineTestCase):
    def setUp(self):
        super().setUp()

        Agreement.objects.invalidate_cache()

    def tearDown(self):
        super().tearDown()

        Agreement.objects.invalidate_cache()

    def test_skip_if_user_is_set(self):
        """pipeline step is skipped if user was passed"""
        request = create_request()
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={},
            backend=backend,
            user=self.user,
            pipeline_index=1,
        )
        self.assertEqual(result, {})

    def test_renders_form_if_not_post(self):
        """pipeline step renders form if not POST"""
        request = create_request()
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        response = create_user_with_form(
            strategy=strategy, details={}, backend=backend, user=None, pipeline_index=1
        )
        self.assertContains(response, "GitHub")

    def test_empty_data_rejected(self):
        """form rejects empty data"""
        request = create_request(data={})
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        response = create_user_with_form(
            strategy=strategy, details={}, backend=backend, user=None, pipeline_index=1
        )
        self.assertEqual(response.status_code, 400)
        self.assertJsonResponseEquals(
            response,
            {
                "email": ["This field is required."],
                "username": ["This field is required."],
            },
        )

    def test_taken_data_rejected(self):
        """form rejects taken data"""
        request = create_request(
            data={"email": self.user.email, "username": self.user.username}
        )
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        response = create_user_with_form(
            strategy=strategy, details={}, backend=backend, user=None, pipeline_index=1
        )
        self.assertEqual(response.status_code, 400)
        self.assertJsonResponseEquals(
            response,
            {
                "email": ["This e-mail address is not available."],
                "username": ["This username is not available."],
            },
        )

    @override_dynamic_settings(account_activation="none")
    def test_user_created_no_activation_verified_email(self):
        """active user is created for verified email and activation disabled"""
        form_data = {"email": "social@auth.com", "username": "SocialUser"}
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={"email": form_data["email"]},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertNewUserIsCorrect(
            new_user, form_data, activation="none", email_verified=True
        )

    @override_dynamic_settings(account_activation="none")
    def test_user_created_no_activation_nonverified_email(self):
        """active user is created for non-verified email and activation disabled"""
        form_data = {"email": "social@auth.com", "username": "SocialUser"}
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={"email": ""},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertNewUserIsCorrect(
            new_user, form_data, activation="none", email_verified=False
        )

    @override_dynamic_settings(account_activation="user")
    def test_user_created_activation_by_user_verified_email(self):
        """active user is created for verified email and activation by user"""
        form_data = {"email": "social@auth.com", "username": "SocialUser"}
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={"email": form_data["email"]},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertNewUserIsCorrect(
            new_user, form_data, activation="user", email_verified=True
        )

    @override_dynamic_settings(account_activation="user")
    def test_user_created_activation_by_user_nonverified_email(self):
        """inactive user is created for non-verified email and activation by user"""
        form_data = {"email": "social@auth.com", "username": "SocialUser"}
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={"email": ""},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertNewUserIsCorrect(
            new_user, form_data, activation="user", email_verified=False
        )

    @override_dynamic_settings(account_activation="admin")
    def test_user_created_activation_by_admin_verified_email(self):
        """inactive user is created for verified email and activation by admin"""
        form_data = {"email": "social@auth.com", "username": "SocialUser"}
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={"email": form_data["email"]},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertNewUserIsCorrect(
            new_user, form_data, activation="admin", email_verified=True
        )

    @override_dynamic_settings(account_activation="admin")
    def test_user_created_activation_by_admin_nonverified_email(self):
        """inactive user is created for non-verified email and activation by admin"""
        form_data = {"email": "social@auth.com", "username": "SocialUser"}
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={"email": ""},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertNewUserIsCorrect(
            new_user, form_data, activation="admin", email_verified=False
        )

    def test_form_check_agreement(self):
        """social register checks agreement"""
        form_data = {"email": "social@auth.com", "username": "SocialUser"}
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        agreement = Agreement.objects.create(
            type=Agreement.TYPE_TOS, text="Lorem ipsum", is_active=True
        )

        response = create_user_with_form(
            strategy=strategy,
            details={"email": form_data["email"]},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        self.assertEqual(response.status_code, 400)
        self.assertJsonResponseEquals(
            response, {"terms_of_service": ["This agreement is required."]}
        )

        # invalid agreement id
        form_data = {
            "email": "social@auth.com",
            "username": "SocialUser",
            "terms_of_service": agreement.id + 1,
        }
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        response = create_user_with_form(
            strategy=strategy,
            details={"email": form_data["email"]},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        self.assertEqual(response.status_code, 400)
        self.assertJsonResponseEquals(
            response, {"terms_of_service": ["This agreement is required."]}
        )

        # valid agreement id
        form_data = {
            "email": "social@auth.com",
            "username": "SocialUser",
            "terms_of_service": agreement.id,
        }
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = create_user_with_form(
            strategy=strategy,
            details={"email": form_data["email"]},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertEqual(new_user.agreements, [agreement.id])
        self.assertEqual(new_user.useragreement_set.count(), 1)

    def test_form_ignore_inactive_agreement(self):
        """social register ignores inactive agreement"""
        form_data = {
            "email": "social@auth.com",
            "username": "SocialUser",
            "terms_of_service": None,
        }
        request = create_request(data=form_data)
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        Agreement.objects.create(
            type=Agreement.TYPE_TOS, text="Lorem ipsum", is_active=False
        )

        result = create_user_with_form(
            strategy=strategy,
            details={"email": form_data["email"]},
            backend=backend,
            user=None,
            pipeline_index=1,
        )

        new_user = User.objects.get(email="social@auth.com")
        self.assertEqual(result, {"user": new_user, "is_new": True})

        self.assertEqual(new_user.agreements, [])
        self.assertEqual(new_user.useragreement_set.count(), 0)


class GetUsernameTests(PipelineTestCase):
    def test_skip_if_user_is_set(self):
        """pipeline step is skipped if user was passed"""
        strategy = create_strategy()
        result = get_username(strategy, {}, None, user=self.user)
        self.assertIsNone(result)

    def test_skip_if_no_names(self):
        """pipeline step is skipped if API returned no names"""
        strategy = create_strategy()
        result = get_username(strategy, {}, None)
        self.assertIsNone(result)

    def test_resolve_to_username(self):
        """pipeline step resolves username"""
        strategy = create_strategy()
        result = get_username(strategy, {"username": "BobBoberson"}, None)
        self.assertEqual(result, {"clean_username": "BobBoberson"})

    def test_normalize_username(self):
        """pipeline step normalizes username"""
        strategy = create_strategy()
        result = get_username(strategy, {"username": "Błop Błoperson"}, None)
        self.assertEqual(result, {"clean_username": "BlopBloperson"})

    def test_resolve_to_first_name(self):
        """pipeline attempts to use first name because username is taken"""
        strategy = create_strategy()
        details = {"username": self.user.username, "first_name": "Błob"}
        result = get_username(strategy, details, None)
        self.assertEqual(result, {"clean_username": "Blob"})

    def test_dont_resolve_to_last_name(self):
        """pipeline will not fallback to last name because username is taken"""
        strategy = create_strategy()
        details = {"username": self.user.username, "last_name": "Błob"}
        result = get_username(strategy, details, None)
        self.assertIsNone(result)

    def test_resolve_to_first_last_name_first_char(self):
        """pipeline will construct username from first name and first char of surname"""
        strategy = create_strategy()
        details = {"first_name": self.user.username, "last_name": "Błob"}
        result = get_username(strategy, details, None)
        self.assertEqual(result, {"clean_username": self.user.username + "B"})

    def test_dont_resolve_to_banned_name(self):
        """pipeline will not resolve to banned name"""
        strategy = create_strategy()
        Ban.objects.create(banned_value="*Admin*", check_type=Ban.USERNAME)
        details = {"username": "Misago Admin", "first_name": "Błob"}
        result = get_username(strategy, details, None)
        self.assertEqual(result, {"clean_username": "Blob"})

    def test_resolve_full_name(self):
        """pipeline will resolve to full name"""
        strategy = create_strategy()
        Ban.objects.create(banned_value="*Admin*", check_type=Ban.USERNAME)
        details = {"username": "Misago Admin", "full_name": "Błob Błopo"}
        result = get_username(strategy, details, None)
        self.assertEqual(result, {"clean_username": "BlobBlopo"})

    def test_resolve_to_cut_name(self):
        """pipeline will resolve cut too long name on second pass"""
        strategy = create_strategy()
        details = {"username": "Abrakadabrapokuskonstantynopolitańczykowianeczkatrzy"}
        result = get_username(strategy, details, None)
        self.assertEqual(result, {"clean_username": "Abrakadabrapok"})


class RequireActivationTests(PipelineTestCase):
    def setUp(self):
        super().setUp()

        self.user.requires_activation = User.ACTIVATION_ADMIN
        self.user.save()

    def test_skip_if_user_not_set(self):
        """pipeline step is skipped if user is not set"""
        request = create_request()
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = require_activation(
            strategy=strategy, details={}, backend=backend, user=None, pipeline_index=1
        )
        self.assertEqual(result, {})

    def test_partial_token_if_user_not_set_no_showstopper(self):
        """pipeline step handles set session token if user is not set"""
        request = create_request()
        strategy = load_strategy(request=request)
        strategy.request.session["partial_pipeline_token"] = "test-token"
        backend = GithubOAuth2(strategy, "/")

        require_activation(
            strategy=strategy, details={}, backend=backend, user=None, pipeline_index=1
        )

    def test_skip_if_user_is_active(self):
        """pipeline step is skipped if user is active"""
        self.user.requires_activation = User.ACTIVATION_NONE
        self.user.save()

        self.assertFalse(self.user.requires_activation)

        request = create_request()
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        result = require_activation(
            strategy=strategy,
            details={},
            backend=backend,
            user=self.user,
            pipeline_index=1,
        )
        self.assertEqual(result, {})

    def test_pipeline_returns_html_response_on_get(self):
        """pipeline step renders http response for GET request and inactive user"""
        request = create_request()
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        response = require_activation(
            strategy=strategy,
            details={},
            backend=backend,
            user=self.user,
            pipeline_index=1,
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "text/html; charset=utf-8")

    def test_pipeline_returns_json_response_on_post(self):
        """pipeline step renders json response for POST request and inactive user"""
        request = create_request(data={"username": "anything"})
        strategy = load_strategy(request=request)
        backend = GithubOAuth2(strategy, "/")

        response = require_activation(
            strategy=strategy,
            details={},
            backend=backend,
            user=self.user,
            pipeline_index=1,
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/json")
        self.assertJsonResponseEquals(
            response,
            {
                "step": "done",
                "backend_name": "GitHub",
                "activation": "admin",
                "email": "test@user.com",
                "username": "TestUser",
            },
        )


class ValidateIpNotBannedTests(PipelineTestCase):
    def test_skip_if_user_not_set(self):
        """pipeline step is skipped if no user was passed"""
        result = validate_ip_not_banned(None, {}, GithubOAuth2)
        self.assertIsNone(result)

    def test_raise_if_banned(self):
        """pipeline raises if user's IP is banned"""
        Ban.objects.create(banned_value="188.*", check_type=Ban.IP)

        try:
            validate_ip_not_banned(
                MockStrategy(user_ip="188.1.2.3"), {}, GithubOAuth2, self.user
            )
            self.fail("validate_ip_not_banned should raise SocialAuthBanned")
        except SocialAuthBanned as e:
            self.assertTrue(isinstance(e.ban, Ban))

    def test_exclude_staff(self):
        """pipeline excludes staff from bans"""
        self.user.is_staff = True
        self.user.save()

        Ban.objects.create(banned_value="188.*", check_type=Ban.IP)

        result = validate_ip_not_banned(
            MockStrategy(user_ip="188.1.2.3"), {}, GithubOAuth2, self.user
        )
        self.assertIsNone(result)


class ValidateUserNotBannedTests(PipelineTestCase):
    def test_skip_if_user_not_set(self):
        """pipeline step is skipped if no user was passed"""
        result = validate_user_not_banned(None, {}, GithubOAuth2)
        self.assertIsNone(result)

    def test_raise_if_banned(self):
        """pipeline raises if user's IP is banned"""
        Ban.objects.create(banned_value=self.user.username, check_type=Ban.USERNAME)

        try:
            validate_user_not_banned(MockStrategy(), {}, GithubOAuth2, self.user)
            self.fail("validate_ip_not_banned should raise SocialAuthBanned")
        except SocialAuthBanned as e:
            self.assertEqual(e.ban.user, self.user)
            self.assertTrue(isinstance(e.ban, BanCache))

    def test_exclude_staff(self):
        """pipeline excludes staff from bans"""
        self.user.is_staff = True
        self.user.save()

        Ban.objects.create(banned_value=self.user.username, check_type=Ban.USERNAME)

        result = validate_user_not_banned(MockStrategy(), {}, GithubOAuth2, self.user)
        self.assertIsNone(result)



from ...threads.models import Thread
from ...threads.permissions import exclude_invisible_threads
from .threads import UserThreads


class UserPosts(UserThreads):
    def get_threads_queryset(self, request, threads_categories, profile):
        return exclude_invisible_threads(
            request.user_acl, threads_categories, Thread.objects
        )

    def get_posts_queryset(self, user, profile, threads_queryset):
        return profile.post_set.select_related("thread", "poster").filter(
            thread_id__in=threads_queryset.values("id")
        )



from django.urls import reverse

from ...test import assert_contains, assert_not_contains
from ..auth import is_admin_authorized

admin_link = reverse("misago:admin:index")


def assert_requires_admin_login(response):
    assert response.status_code == 200
    assert_contains(response, "Administration")
    assert_contains(response, "Sign in")


def test_anonymous_user_is_asked_to_login_to_access_admin_view(db, client):
    response = client.get(admin_link)
    assert_requires_admin_login(response)


def test_authenticated_user_is_asked_to_login_to_access_admin_view(client, user):
    client.force_login(user)
    response = client.get(admin_link)
    assert_requires_admin_login(response)


def test_unathorized_admin_is_asked_to_login_to_access_admin_view(client, superuser):
    client.force_login(superuser)
    response = client.get(admin_link)
    assert_requires_admin_login(response)


def test_authorized_admin_is_allowed_to_access_admin_view(admin_client):
    response = admin_client.get(admin_link)
    assert is_admin_authorized(response.wsgi_request)
    assert_not_contains(response, "Sign in")


def test_admin_authorization_is_checked_on_admin_view_access(mocker, client, user):
    admin_authorization = mocker.patch(
        "misago.admin.middleware.is_admin_authorized", return_value=False
    )
    response = client.get(admin_link)
    admin_authorization.assert_called_once_with(response.wsgi_request)



# Generated by Django 2.2.1 on 2019-05-18 16:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("misago_themes", "0002_create_default_theme_and_cache_version")]

    operations = [
        migrations.AlterField(
            model_name="theme",
            name="level",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="theme",
            name="lft",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="theme",
            name="rght",
            field=models.PositiveIntegerField(editable=False),
        ),
    ]



from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from ....admin.views import generic
from ...datadownloads import (
    expire_user_data_download,
    request_user_data_download,
    user_has_data_download_request,
)
from ...models import DataDownload
from ..forms import RequestDataDownloadsForm, FilterDataDownloadsForm


class DataDownloadAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:users:data-downloads:index"
    templates_dir = "misago/admin/datadownloads"
    model = DataDownload


class DataDownloadsList(DataDownloadAdmin, generic.ListView):
    items_per_page = 30
    ordering = [("-id", _("From newest")), ("id", _("From oldest"))]
    selection_label = _("With data downloads: 0")
    empty_selection_label = _("Select data downloads")
    mass_actions = [
        {
            "action": "expire",
            "name": _("Expire downloads"),
            "confirmation": _(
                "Are you sure you want to set selected data downloads as expired?"
            ),
        },
        {
            "action": "delete",
            "name": _("Delete downloads"),
            "confirmation": _(
                "Are you sure you want to delete selected data downloads?"
            ),
        },
    ]
    filter_form = FilterDataDownloadsForm

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related("user", "requester")

    def action_expire(self, request, data_downloads):
        for data_download in data_downloads:
            expire_user_data_download(data_download)

        messages.success(
            request, _("Selected data downloads have been set as expired.")
        )

    def action_delete(self, request, data_downloads):
        for data_download in data_downloads:
            data_download.delete()

        messages.success(request, _("Selected data downloads have been deleted."))


class RequestDataDownloads(DataDownloadAdmin, generic.FormView):
    form_class = RequestDataDownloadsForm

    def handle_form(self, form, request):
        for user in form.cleaned_data["users"]:
            if not user_has_data_download_request(user):
                request_user_data_download(user, requester=request.user)

        messages.success(
            request, _("Data downloads have been requested for specified users.")
        )



from django.core.exceptions import PermissionDenied
from social_django.strategy import DjangoStrategy


class MisagoStrategy(DjangoStrategy):
    def setting(self, name, default=None, backend=None):
        if backend:
            backend_settings = self.request.socialauth[backend.name]["settings"]
            if name in backend_settings:
                return backend_settings[name]

        return super().setting(name, default, backend)

    def authenticate(self, backend, *args, **kwargs):
        kwargs["strategy"] = self
        kwargs["storage"] = self.storage
        kwargs["backend"] = backend

        try:
            return backend.authenticate(*args, **kwargs)
        except PermissionDenied:
            pass



from rest_framework import serializers

from ..threads.validators import validate_post_length


class MarkupSerializer(serializers.Serializer):
    post = serializers.CharField(required=False, allow_blank=True)

    def validate(self, data):
        settings = self.context["settings"]
        validate_post_length(settings, data.get("post", ""))
        return data



from django.urls import reverse

from ....test import assert_has_error_message


def test_exporting_default_theme_sets_error_message(admin_client, default_theme):
    export_link = reverse("misago:admin:themes:export", kwargs={"pk": default_theme.pk})
    response = admin_client.post(export_link)
    assert_has_error_message(response)


def test_exporting_nonexisting_theme_sets_error_message(
    admin_client, nonexisting_theme
):
    export_link = reverse(
        "misago:admin:themes:export", kwargs={"pk": nonexisting_theme.pk}
    )
    response = admin_client.post(export_link)
    assert_has_error_message(response)



from ..models import Agreement


def set_agreement_as_active(agreement, commit=False):
    agreement.is_active = True
    Agreement.objects.filter(type=agreement.type).exclude(pk=agreement.pk).update(
        is_active=False
    )

    if commit:
        agreement.save(update_fields=["is_active"])
        Agreement.objects.invalidate_cache()


def disable_agreement(agreement, commit=False):
    agreement.is_active = False
    if commit:
        agreement.save(update_fields=["is_active"])
        Agreement.objects.invalidate_cache()



from django.http import Http404
from django.test import TestCase, override_settings
from django.urls import reverse

from ..shortcuts import get_int_or_404


@override_settings(ROOT_URLCONF="misago.core.testproject.urls")
class PaginateTests(TestCase):
    def test_valid_page_handling(self):
        """Valid page number causes no errors"""
        response = self.client.get(reverse("test-pagination", kwargs={"page": 2}))
        self.assertEqual("5,6,7,8,9", response.content.decode())

    def test_invalid_page_handling(self):
        """Invalid page number results in 404 error"""
        response = self.client.get(reverse("test-pagination", kwargs={"page": 42}))
        self.assertEqual(response.status_code, 404)

    def test_implicit_page_handling(self):
        """Implicit page number causes no errors"""
        response = self.client.get(reverse("test-pagination"))
        self.assertEqual("0,1,2,3,4", response.content.decode())

    def test_explicit_page_handling(self):
        """Explicit page number results in redirect"""
        response = self.client.get(reverse("test-pagination", kwargs={"page": 1}))
        valid_url = "/forum/test-pagination/"
        self.assertEqual(response["Location"], valid_url)


@override_settings(ROOT_URLCONF="misago.core.testproject.urls")
class ValidateSlugTests(TestCase):
    def test_valid_slug_handling(self):
        """Valid slug causes no interruption in view processing"""
        response = self.client.get(
            reverse("validate-slug-view", kwargs={"slug": "eric-the-fish", "pk": 1})
        )
        self.assertContains(response, "Allright")

    def test_invalid_slug_handling(self):
        """Invalid slug returns in redirect to valid page"""
        response = self.client.get(
            reverse("validate-slug-view", kwargs={"slug": "lion-the-eric", "pk": 1})
        )

        valid_url = "/forum/test-valid-slug/eric-the-fish-1/"
        self.assertEqual(response["Location"], valid_url)


class GetIntOr404Tests(TestCase):
    def test_valid_inputs(self):
        """get_int_or_404 returns int for valid values"""
        VALID_VALUES = [("0", 0), ("123", 123), ("000123", 123), ("1", 1)]

        for value, result in VALID_VALUES:
            self.assertEqual(get_int_or_404(value), result)

    def test_invalid_inputs(self):
        """get_int_or_404 raises Http404 for invalid values"""
        INVALID_VALUES = [
            None,
            "",
            "str",
            "5tr",
            "s0tr",
            "str123",
            "12.321",
            ".4",
            "5.",
        ]

        for value in INVALID_VALUES:
            with self.assertRaises(Http404):
                get_int_or_404(value)


@override_settings(ROOT_URLCONF="misago.core.testproject.urls")
class PaginatedResponseTests(TestCase):
    def test_page_response(self):
        """utility returns response for only page arg"""
        response = self.client.get(reverse("test-paginated-response"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "results": [i + 10 for i in range(10)],
                "page": 2,
                "pages": 10,
                "count": 100,
                "first": 1,
                "previous": 1,
                "next": 3,
                "last": 10,
                "before": 10,
                "more": 80,
            },
        )

    def test_explicit_data_response(self):
        """utility returns response with explicit data"""
        response = self.client.get(reverse("test-paginated-response-data"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "results": ["a", "b", "c", "d", "e"],
                "page": 2,
                "pages": 10,
                "count": 100,
                "first": 1,
                "previous": 1,
                "next": 3,
                "last": 10,
                "before": 10,
                "more": 80,
            },
        )

    def test_explicit_serializer_response(self):
        """utility returns response with data serialized via serializer"""
        response = self.client.get(reverse("test-paginated-response-serializer"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "results": [{"id": 0}, {"id": 2}, {"id": 4}, {"id": 6}],
                "page": 1,
                "pages": 1,
                "count": 4,
                "first": None,
                "previous": None,
                "next": None,
                "last": None,
                "before": 0,
                "more": 0,
            },
        )

    def test_explicit_data_serializer_response(self):
        """utility returns response with explicit data serialized via serializer"""
        response = self.client.get(reverse("test-paginated-response-data-serializer"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "results": [{"id": "aa"}, {"id": "bb"}, {"id": "cc"}, {"id": "dd"}],
                "page": 1,
                "pages": 1,
                "count": 4,
                "first": None,
                "previous": None,
                "next": None,
                "last": None,
                "before": 0,
                "more": 0,
            },
        )

    def test_explicit_data_extra_response(self):
        """utility returns response with explicit data and extra"""
        response = self.client.get(reverse("test-paginated-response-data-extra"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "results": ["a", "b", "c", "d"],
                "page": 1,
                "pages": 1,
                "count": 4,
                "first": None,
                "previous": None,
                "next": "EXTRA",
                "last": None,
                "before": 0,
                "more": 0,
                "lorem": "ipsum",
            },
        )



from django.contrib.auth import get_user_model

from . import PostingEndpoint, PostingMiddleware
from ...models import Subscription

User = get_user_model()


class SubscribeMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        return self.mode != PostingEndpoint.EDIT

    def post_save(self, serializer):
        self.subscribe_new_thread()
        self.subscribe_replied_thread()

    def subscribe_new_thread(self):
        if self.mode != PostingEndpoint.START:
            return

        if self.user.subscribe_to_started_threads == User.SUBSCRIPTION_NONE:
            return

        self.user.subscription_set.create(
            category=self.thread.category,
            thread=self.thread,
            send_email=self.user.subscribe_to_started_threads == User.SUBSCRIPTION_ALL,
        )

    def subscribe_replied_thread(self):
        if self.mode != PostingEndpoint.REPLY:
            return

        if self.user.subscribe_to_replied_threads == User.SUBSCRIPTION_NONE:
            return

        try:
            return self.user.subscription_set.get(thread=self.thread)
        except Subscription.DoesNotExist:
            pass

        # posts user's posts in this thread, minus events and current post
        posts_queryset = self.user.post_set.filter(
            thread=self.thread, is_event=False
        ).exclude(pk=self.post.pk)

        if posts_queryset.exists():
            return

        self.user.subscription_set.create(
            category=self.thread.category,
            thread=self.thread,
            send_email=self.user.subscribe_to_replied_threads == User.SUBSCRIPTION_ALL,
        )
